from junkyard.garbagebasket.PrintHandler import c_print


def collapse_list(lst, silent=False):
    """
    function collapses lists into flat strings
    example: collapse_list(['a', 'b', ' ', 'c'])
    input: ['a', 'B', ' ', 'c']
    output: aBc

    :param lst: input list to be collapsed into a single string of characters
    :param silent: toggle silence of error reporting
    :return: collapsed-to-string version of the input list
    """
    function_name = ""
    if isinstance(lst, list):
        o_str = str(lst)
        filters = ["[", "]", ",", "'", " "]
        for f in filters:
            o_str = o_str.replace(f, "")
        return o_str
    else:
        c_print(f"{function_name}: INVALID INPUT", color="RED", silent=silent)


def handle_errors(error, log=True, silent=False, msg=None):
    def error_msg(_msg):
        """ sub function to do error printing """
        if log:
            c_print(_msg, log_mode=4, silent=silent)
        else:
            c_print(_msg, color="RED", silent=silent)

    function_name = "alt_handler"
    handled = False
    errors = {"FileNotFound": f"FileNotFoundError({msg})",
              "PermissionError": f"PermissionError({msg})",
              "FileExistsError": f"FileNotFoundError({msg})",
              "OSError": f"OSError({msg})",
              "IndexError": f"IndexError({msg})",
              "ModeError": f"ModeOutOfRangeError({msg})"}
    for kw in errors:
        if error == kw:
            value = errors.get(kw)
            error_msg(value)
            handled = True
            break

    if not handled:
        c_print(f"UNKNOWN ERROR WENT UNHANDLED", color="RED")



# import copy
import hashlib
import os
import random
# import requests
import shutil

# get tkinter for file dialogs
from tkinter import filedialog

# get downloading backbone module wget
import wget

# import other junkyard stuff
from junkyard.garbagebasket.PrintHandler import init as print_init
from junkyard.garbagebasket.PrintHandler import c_print, sl_print
from junkyard.garbagebasket.SystemHandler import handle_errors, collapse_list

handler_name = "FileHandler"

# initialize all of the outer scope variables
# standard read/write/append modes
write_modes = ("w", "write", "Write", "WRITE")
write_append_modes = ("a", "append", "Append", "APPEND")
read_modes = ("r", "read", "Read", "READ")
# TODO: do we need this?
#       do we want to add the r+ / rb+ functionality, or do the current
#       tools suffice?
read_write_modes = ("r+", "R+", "rw", "RW", "readwrite", "READWRITE")

# bytes modes
write_bytes_modes = ("wb", "WB", "w_bytes", "W_bytes")
write_append_bytes_modes = ("ab", "AB", "a_bytes", "A_bytes")
read_bytes_modes = ("rb", "RB", "r_bytes", "R_bytes")
# TODO: do we need this?
#       do we want to add the r+ / rb+ / r+b functionality, or do the current
#       tools suffice?
read_write_bytes_modes = ("rb+", "RB+", "rw_bytes", "RW_bytes")
# TODO: What about other modes like "x"?

silence = False
logging = False


def init(log=False, silent=False):
    """
    WORK IN PROGRESS

    FileHandler initializer function.
    while not required, it can be used to toggle the global variable "silence"
    which will disable all print statements within the functions

    *all functions in file can be silenced with silent=True*
    * BUT OTHERWISE HAVE INFORMATIVE PRINT STATEMENTS *

    :param log:
    :param silent: a True or False bool to determine whether or not to send print statements to console
    """
    global silence
    global logging

    if not silent and not silence:
        # do a sys.stdout.write and flush at top and bottom to print on same line
        sl_print("FileHandler() Initializing")

    silence = silent
    logging = log

    if logging:
        print_init(0)
    else:
        print_init(5)

    # if silent is false
    if not silent and not silence:
        # second stdout.write and flush to clear old message
        # and replace it with a new one
        sl_print("FileHandler() Initialized\n")


def is_directory(directory, silent=False):
    # setup help() return for the function
    """
    function to return True or False as to whether a directory exists

    :param directory: path to the directory the function will check for

    :param silent: a True or False bool to determine whether or not to make print statements

    :return: has 2 possible returns (True, False)
             True = directory exists
             False = directory does not exist
    """

    # establish function name for print/debug statements
    function_name = f"{handler_name}.is_directory()"
    silent = True if silence else silent

    # load the return of os.path.isdir() into a variable
    # to later be returned as True or False
    try:
        exists = os.path.isdir(directory)
    # account for invalid input in directory
    except TypeError:
        handle_errors(error="TypeError", log=logging, silent=silent,
                      msg=f"{function_name}: ({os.getcwd()}\\<{directory}>) triggered a TypeError in is_directory() "
                          f"- check the input")
        exists = False

    # and the directory exists
    if exists:
        # say found
        c_print(f"{function_name}: ({os.getcwd()}\\<{directory}>) found", color="GREEN", silent=silent)
    else:
        # say 'I'm sorry for the loss' -- I mean -- "not found"
        c_print(f"{function_name}: ({os.getcwd()}\\<{directory}>) not found", color="L_RED", silent=silent)
    # return the True False nature of the exists variable
    return exists


def is_file(filename, silent=False):
    # setup help() return for the function
    """
    function to return True or False as to whether a file exists

    :param filename: path to file the function will check for

    :param silent: a True or False bool to determine whether or not to make print statements

    :return: has 2 possible returns (True, False)
             True = file exists
             False = file does not exist
    """

    # establish function name for print/debug statements
    function_name = f"{handler_name}.is_file()"
    silent = True if silence else silent
    try:
        # load the True/False os.path.isfile() return into the exists variable
        # to later be returned as a file exists or file doesnt exist
        exists = os.path.isfile(filename)

        # account for invalid input in directory

    except TypeError:
        handle_errors(error="TypeError", log=logging, silent=silent,
                      msg=f"{function_name}: ({os.getcwd()}\\<{filename}>) triggered a TypeError in is_file() "
                          f"- check the input")
        exists = False

    # if file exists
    if exists:
        # say found
        c_print(f"{function_name}: ({os.getcwd()}\\<{filename}>) found", color="GREEN", silent=silent)
    # if file doesn't exist
    else:
        # say 'I'm sorry for the loss' -- I mean -- "not found"
        c_print(f"{function_name}: ({os.getcwd()}\\<{filename}>) not found", color="L_RED", silent=silent)
    # return the True False nature of the exists variable
    return exists


def file_verification(file_1, file_2, silent=False):
    """
    verifies that two files are the same file to verify file duplication success

    :param file_1:
    :param file_2:
    :param silent:
    :return:
    """
    function_name = f"{handler_name}.file_verification()"
    silent = True if silence else silent

    def hash_bytestr_iter(bytesiter, hasher, ashexstr=False):
        for block in bytesiter:
            hasher.update(block)
        return hasher.hexdigest() if ashexstr else hasher.digest()

    def file_as_blockiter(afile, blocksize=65536):
        with afile:
            block = afile.read(blocksize)
            while len(block) > 0:
                yield block
                block = afile.read(blocksize)

    if is_file(file_1, silent=True) and is_file(file_2, silent=True):
        f_name_list = [file_1, file_2]
        file_1_hash, file_2_hash = [(f_name, hash_bytestr_iter(file_as_blockiter(open(f_name, 'rb')), hashlib.sha256()))
                                    for f_name in f_name_list]

        match = True if file_1_hash[1] == file_2_hash[1] else False
        color = "Green" if match else "L_RED"
        c_print(f"{function_name}: VERIFYING HASHES", color=color, silent=silent)
        c_print(file_1_hash[1], color=color, silent=silent)
        c_print(file_2_hash[1], color=color, silent=silent)

        # ternary conditional return, if the hashes match return true
        # and if the hashes don't match return false
        return True if match else False

    else:
        c_print(f"{function_name}: ONE OR MORE FILES DO NOT EXIST", color="L_RED", silent=silent)
        return False


# establish function name for print/debug statements
def generate_temp_filename(f_name, r_length=16, silent=False):
    # list of random characters
    function_name = f"{handler_name}.generate_temp_filename()"
    silent = True if silence else silent

    choices = ("A", "B", "C", "D", "E", "F", "G", "H",
               "I", "J", "K", "L", "M", "N", "O", "P",
               "Q", "R", "S", "T", "U", "V", "W", "X",
               "Y", "Z", "0", "1", "2", "3", "4", "5",
               "6", "7", "8", "9")

    # generate the temp suffix
    r_suffix = [random.choice(choices) for x in range(r_length) if x is not None]
    # strip f_name list
    r_suffix = collapse_list(r_suffix)

    # create temporary file
    try:
        f_name = uniform_filename(f_name)
        extension = f".{f_name.split('.')[-1]}" if "." in f_name else ""
        temp_file = f"{f_name.split('.')[0]}{r_suffix}{extension}'" if "." in f_name else f"{f_name}{r_suffix}"
        temp_file = uniform_filename(temp_file)
        return temp_file
    except IndexError:
        handle_errors(error="IndexError", log=logging, silent=silent,
                      msg=f"{function_name}: could not generate full temporary file name, "
                          f"returning random suffix instead")
        return r_suffix


def uniform_filename(filename):
    """
    function makes filename directory strings uniform to the "\\" backslash standard

    :param filename: file directory string you want to format to a "\\" standard
    :return:
    """
    # function_name = f"{handler_name}.convert_filename()"
    filters = ["/", "//", "\\\\"]
    for f in filters:
        if f in filename:
            filename = filename.replace(f, "\\")
    return filename


def split_filename(filename, silent=False):
    # setup help() return for the function
    """
    function to split directories and file names out of file paths and return both directory and file name

    :param filename: path to the file the function will split the filename and directory from
    :param silent: a True or False bool to determine whether or not to make print statements
    :return: returns the post split filename on index[0] and directory on index[1]
    """

    # establish function name for print/debug statements
    function_name = f"{handler_name}.split_filename()"
    silent = True if silence else silent
    # if the filename contains "/"s which are a good indicator
    # that they hold a directory path
    if "/" in filename:
        # split the directory into a list separated by /'s
        new_directory = filename.split("/")
        # create a [:-value] location for the beginning of the actual filename
        # by subtracting the length of the last split against he original length
        # of the filename
        extract_index = len(filename) - len(new_directory[-1])
        directory = filename[:extract_index]
        # declare the new filename as the file index in new_directory
        filename = new_directory[-1]

    # other wise there is no directory hidden in filename, and we can just leave things as they are
    else:
        # and assign a None to directory
        directory = None

    # if not silent
    # we'll print out the returns
    c_print(f"{function_name}:", silent=silent)
    c_print(f"working directory: {os.getcwd()}", silent=silent)
    c_print(f"directory: {directory}", silent=silent)
    c_print(f"filename: {directory}", silent=silent)

    # and/or return the filename, and directory
    return filename, directory


def touch_file(filename, directory=None, mode=None, silent=False):
    # setup help() return for the function
    """
    function to create a file (and associated directories if needed) if it does
    not already exist.

    :param filename:  path to the file (may include directory) the function will create

    :param directory: OPTIONAL path to the directory containing the file

    :param mode: has 2 possible modes "w" and "wb"
                 "w" to touch a plain-text file into existence
                 "wb" to touch a bytes file into existence


    :param silent: a True or False bool to determine whether or not to make print statements

    :return: has 2 possible returns (True, False)
             True: infers that the file exists and that we were otherwise successful
             False: infers that the file not only did not exists, but continues not existing
                    and that the function has failed

    """

    # establish function name for print/debug statements
    function_name = f"{handler_name}.touch_file()"
    silent = True if silence else silent
    # check if there is a value in the directory field
    # if not we want to double check against the filename
    if directory is None:
        # call the split_filename() function and check index 0 for a filename
        # and index 1 for a directory and load them into temporary variables
        # to preserve the original filename
        temp_filename, temp_directory = split_filename(filename, silent=True)

        # we make sure that the directory that split_filename() function
        # returned isn't None, meaning it has a value, and therefore the
        # original filename had a directory in it.
        if temp_directory is not None:
            # make sure the filename is updated to reflect the split
            filename = temp_filename
            # make sure the directory is updated to reflect the split
            directory = temp_directory

            # next we want to make sure that the directory does or does not already exist
            # and create it if it doesn't using the make_directory() function
            make_directory(directory, silent=True)

        # no hidden directory so pass out to next phase
        else:
            pass

    # if there is a value in the field
    else:
        # if the directory doesn't exist, create it
        # the make_directory() function automatically checks
        # check if the directory exists first
        make_directory(directory, silent=True)

    # if there is some form of value in directory at this point
    # we should append the directory and filename together
    # at this point either way they should be split
    if directory is not None:
        filename = f"{directory}{filename}"

    # set the proper modes and content string type based on mode field argument
    # write
    if mode in write_modes:
        mode = "w"
        blank_content = ""

    # write bytes
    elif mode in write_bytes_modes:
        mode = "wb"
        # use a bytes string for write bytes
        blank_content = b""

    # if mode is None or w/e just default to write mode
    else:
        mode = "w"
        blank_content = ""

    # if the file doesn't exist
    if not is_file(filename, silent=True):
        # protect from errors -- shouldn't get them -- but we're protected if we do.
        try:
            # open it with a context manager as f and write an empty blank line based on the selected mode
            # inserting the respective string type as blank_content
            # - the context manager will automatically close the file for us <3
            with open(filename, mode) as f:
                f.write(blank_content)
        # take care of all those pesky potential errors due to potential user error
        # and return false to declare function failure
        except ValueError:
            handle_errors(error="ValueError", log=logging, silent=silent,
                          msg=f"{function_name}: ValueError with ('{filename}') as the filename, "
                              f"running in '{mode}' mode")
            return False

        except TypeError:
            handle_errors(error="TypeError", log=logging, silent=silent,
                          msg=f"{function_name}: TypeError with ('{filename}') as the filename, "
                              f"running in '{mode}' mode")
            return False

        except FileNotFoundError:
            handle_errors(error="FileNotFoundError", log=logging, silent=silent,
                          msg=f"{function_name}: FileNotFoundError with  ('{filename}') as the filename, "
                              f"running in '{mode}' mode")
            return False

        # load the bool of is_file() into a variable to determine whether or not we failed to touch the file
        exists = is_file(filename, silent=True)
        # if the file exists, we succeeded
        if exists:
            # if we're chatty cathy
            if not silent and not silence:
                # say "we rule, the genes are on top of shit, we won, everything is good and nothing is bad"
                c_print(f"{function_name}: successfully touched ({os.getcwd()}\\<{filename}>)", color="GREEN")

        # if the file doesn't exist
        else:
            # and we're talking
            if not silent and not silence:
                # say "we suck, and failed a simple task... that we are failures -- just like the ancestors"
                # let them know that this failure is not only a programmatic failure but also a genetic one
                c_print(f"{function_name}: failed to touch ({os.getcwd()}\\<{filename}>)", color="RED")

        # and/or return the exists bool to represent success or failure
        return exists

    # if the file already exists, skip the whole song and dance and write home about it
    else:
        # if we can talk
        if not silent and not silence:
            # we will talk, we'll tell them how their precious file already exists
            c_print(f"{function_name}: ({os.getcwd()}\\<{filename}>) already exists", color="YELLOW")

            # we will return a True value because the file existing, is functionally
            # the same as successfully touching the file
            return True


def make_directory(directory, silent=False):
    # setup help() return for the function
    """
    function to CREATE NEW DIRECTORIES if they don't already exists

    :param directory: path to the directory the function is creating

    :param silent: a True or False bool to determine whether or not to make print statements

    :return: has 2 possible return values of True and False (bool)
             True: The directory was successfully created, or otherwise already exists
                   both are functionally the same
             False: The directory does not exist, or the function was otherwise unable to
                    create the directory
    """

    # establish function name for print/debug statements
    function_name = f"{handler_name}.make_directory()"
    silent = True if silence else silent
    # if the file exists pass
    if is_directory(directory, silent=True):
        # if silent == false
        if not silent and not silence:
            # let them know that the directory exists and that there is no need
            # to create it
            c_print(f"{function_name}: could not create the directory ({os.getcwd()}\\<{directory}>)")
            c_print("because it already exists")

        # we'll return with a True, to say that YES the directory is there
        # even though *this function* didn't create it.
        # the goal of having the directory exist is still met.
        return True

    # otherwise create the directory
    else:
        # attempt to create the directory using os.makedirs()
        try:
            os.makedirs(directory)
        # if TypeError
        except TypeError:
            handle_errors(error="TypeError", log=logging, silent=silent,
                          msg=f"{function_name}: ({os.getcwd()}\\<{directory}>) triggered a TypeError in "
                          f"make_directory({directory}) -- check input")
            return False

        # if ValueError
        except ValueError:
            handle_errors(error="ValueError", log=logging, silent=silent,
                          msg=f"{function_name}: ({os.getcwd()}\\<{directory}>) triggered a ValueError in"
                          f" make_directory({directory}) -- check input")
            # return False to state that the function failed
            return False
        # check if the directories were successfully created and return the bool into a variable
        exists = is_directory(directory, silent)
        # use the exists bool to determine how we continue
        # if directories were successfully created
        if exists:
            c_print(f"{function_name}: ({os.getcwd()}\\<{directory}>) was successfully created!",
                    color="GREEN", silent=silent)
        # if directories were unsuccessfully created
        else:
            # let them know that it was a failure
            c_print(f"{function_name}: failed to create ({os.getcwd()}\\<{directory}>) as a directory",
                    color="RED", silent=silent)

        # return the bool to determine success or failure
        return exists


def read_file(filename, mode=None, silent=False):
    # setup help() return for the function
    """
    function to READ return the contents of a file

    :param filename: path to the file the function is reading

    :param mode: has 3 possible modes ('r', 'rb', 'None')
                 'None' - Attempts to decide the read mode automatically (Experimental)
                 'r'    - reads the file in 'r' (plain text) mode
                 'rb'   - reads the file in 'rb' (bytes) mode

    :param silent: a True or False bool to determine whether or not to make print statements

    :return: has 2 possible returns (Content, None)
             the function should return the content of the file
             or if an error is encountered along the way
             it will return None
    """

    # establish function name for print/debug statements
    function_name = f"{handler_name}.read_file()"
    silent = True if silence else silent
    if is_file(filename, silent=True):
        # set mode accordingly
        # TODO: Investigate additional read modes
        # read plain
        if mode in read_modes:
            mode = "r"
        # read bytes
        elif mode in read_bytes_modes:
            mode = "rb"

        # if the mode is None / unfilled
        elif mode is None:
            #       INCLUDE AUTOMATIC PROCESSING DURING "mode=None" CALLS
            #       -- ENDED UP TAKING A DIFFERENT APPROACH, WE USE TRY/EXCEPT TYPEERROR
            #          TO CATCH IF WE HAVE THE WRONG MODE, IF WE DO WE TRY THE OTHER READ
            #          MODE.
            #          HOPING THIS WORKS!
            #       -- (SEEMS TO BE WORKING)

            try:
                with open(filename, "r") as f:
                    return f.read()
            except TypeError:
                with open(filename, "rb") as f:
                    return f.read()

        # use a context manager to open the file and return its contents
        # the context manager will automatically close the file once
        try:
            with open(filename, mode) as f:
                return f.read()
        # handle pesky errors that may arise for whatever reason and inform the user
        # and then pass out of the function, returning None
        except ValueError:
            handle_errors(error="ValueError", log=logging, silent=silent,
                          msg=f"{function_name}: ValueError with ('{filename}') as the filename, "
                              f"running in '{mode}' mode")
            return None
        except TypeError:
            handle_errors(error="TypeError", log=logging, silent=silent,
                          msg=f"{function_name}: TypeError with ('{filename}') as the filename, "
                              f"running in '{mode}' mode")
            return None

    # if the file doesn't
    else:
        # if we can tell secrets
        # announce it to the world
        c_print(f"{function_name}: file not found.", color="L_RED", silent=silent)
        # and return None to represent the file doesn't exist
        return None


# TODO: INCREASE SAFETY OF THE WRITE FUNCTION BY USING A TEMP FILE (DONE)
def write_file(content, filename, mode=None, encoding='utf-8', overwrite=False, verify_write=True, silent=False):
    # setup help() return for the function
    """
    function to WRITE FILES

    :param content: content to write to the file

    :param filename: path to the file the function is writing

    :param mode: write mode, comes in two 5 flavors, write, writebytes, append, appendbytes, and None
                 None(Default) mode - will determine if the file exists and make it if it doesn't
                                      and determine from the content type whether or not to write
                                      in bytes or plain modes.  --  this is essentially an automatic
                                      mode.

                 "w" (write) mode - will write the file as a plain text with a new line at the end
                                    !CAUTION! - THIS WILL WIPE THE FILE'S PREVIOUS CONTENTS, IF ANY
                                                AND REPLACE IT WITH THE CONTENTS OF THE CONTENT FIELD

                 "wb" (write bytes) mode - this mode is a lot like "w" mode except it writes in bytes
                                           without a new line, but carries the same content wiping
                                           warning as "w" mode.

                 "a" (append) mode - this mode should be used on existing files, and will add the "content" field
                                     to the last line of a plain text type file (such as .txt)

                 "ab" (append bytes) mode - this mode is meant to add bytes type content directly to the end
                                            of a bytes file (such as .jpg, or other data type) :shrug:

    :param encoding: encoding used for bytes mode -- example usage: 'utf-8'

    :param silent: a True or False bool to determine whether or not to make print statements

    :param overwrite: whether to overwrite or not

    :param verify_write: check md5 checksum of temporary file and target file before removing the temporary file

    :return: (True, False) bool return to confirm that the file exists at end of function
             if it returns false it means that the write has somehow failed.
    """

    function_name = f"{handler_name}.write_file()"
    silent = True if silence else silent

    # we'll set the modes to their proper values based on the user's call
    # and assign the proper content type
    # set mode to standard write
    if mode in write_modes:
        mode = "w"
        # create a new line each entry
        content = f'{content}\n'

    # set mode to write bytes
    elif mode in write_bytes_modes:
        mode = "wb"
        # make sure the data is bytes type and give it an encoding
        if isinstance(content, bytes):
            pass
        else:
            content = bytes(content, encoding=encoding)

    # set mode to append standard
    elif mode in write_append_modes:
        mode = "a"
        # create a new line each entry
        content = f'{content}\n'

    # set mode to append bytes
    elif mode in write_append_bytes_modes:
        mode = "ab"
        # make sure the data is bytes type and give it an encoding
        if isinstance(content, bytes):
            pass
        else:
            content = bytes(content, encoding=encoding)

    # if the mode field argument is empty
    if mode is None:
        # if the file exists
        if is_file(filename, silent=True):
            # and is_bytes is set to zero
            if isinstance(content, bytes):
                # set to append bytes
                mode = "ab"
            # if the content isn't bytes
            else:
                # set to append (plain txt)
                mode = "a"
                # set content to string() type
                content = f'{content}\n'
        # if the file doesnt exist
        else:
            # and is_bytes is set to zero
            if isinstance(content, bytes):
                # set to write bytes
                mode = "wb"
            # if is_bytes is anything but 0
            else:
                # set to write (plain txt)
                mode = "w"
                # set content to string() type
                content = f'{content}\n'

    # let's make sure that we don't get any errors
    # TODO: MAKE SAFE WRITE WITH TEMP FILE (DONE, refine?) (UPDATE COMMENTS!)
    temporary_file = generate_temp_filename(filename)
    try:
        # create a temporary file and if in append mode
        # copy content of existing file (if it exists) into the temporary file
        # and then write content to the temporary file

        # for appending
        # if the file exists copy it as the base for the temporary file
        if is_file(filename, silent=True):
            if mode in ("a", "ab"):
                copy_file(filename, temporary_file, mode="copy2", overwrite=overwrite, silent=True)

        # write the content to the temp file (append if in append mode)
        with open(temporary_file, mode) as outfile:
            outfile.write(content)

    # such as value errors or...
    except ValueError:
        handle_errors(error="ValueError", log=logging, silent=silent,
                      msg=f"{function_name}: ValueError with ('{filename}') as the filename, running in '{mode}' mode")
        return False
    # type errors
    except TypeError:
        handle_errors(error="TypeError", log=logging, silent=silent,
                      msg=f"{function_name}: TypeError with ('{filename}') as the filename, running in '{mode}' mode")
        return False
    # permission errors
    except PermissionError:
        handle_errors(error="PermissionError", log=logging, silent=silent,
                      msg=f"{function_name}: PermissionError with ('{filename}') as the filename, "
                          f"running in '{mode}' mode")
        return False

    # after we finish the temp_file write we can return whether or not the file
    # was successfully created using the is_file() function
    temp_exists = is_file(temporary_file, silent=True)
    if temp_exists:
        # check to see if the current file already exists
        # delete the original file
        # copy the temporary_file to the filename
        copy_file(temporary_file, filename, mode="copy2", overwrite=True, silent=True)
        # remove the temporary file if it and the target file exist
        if is_file(temporary_file, silent=True) and is_file(filename, silent=True):
            # if verify_write is true then get the sha256 returns of both the temporary file
            # and the target file and compare them, if they are not the same, do not delete
            # the temporary file
            if verify_write:
                # TODO: Verification Function (DONE)
                if file_verification(temporary_file, filename, silent=True):
                    remove_file(temporary_file, silent=True)
                    if is_file(temporary_file, silent=True):
                        c_print(f"{function_name}: failed to automatically remove the temporary file",
                                color="L_RED", silent=silent)
                    else:
                        c_print(f"{function_name}: successfully removed the temporary file",
                                color="GREEN", silent=silent)

                else:
                    c_print(f"{function_name}: failed to write file.", color="RED", silent=silent)
                    c_print(f"your data can be recovered from {temporary_file}", color="L_RED", silent=silent)
                    return False
            else:
                remove_file(temporary_file, silent=True)
                if is_file(temporary_file, silent=True):
                    c_print(f"{function_name}: failed to automatically remove the temporary file",
                            color="RED", silent=silent)
                    c_print(f"{split_filename(temporary_file)[1]}", color="L_RED", silent=silent)
                    return False
                return True
        else:
            c_print(f"{function_name}: failed to write the temporary file file.", color="RED", silent=silent)
            return False

        # after we finish the write we can return whether or not the file
        # was successfully created using the is_file() function
        exists = is_file(filename, silent=True)

        # if the file exists
        if exists:
            c_print(f"{function_name}: successfully wrote {filename} in '{mode}' mode",
                    color="GREEN", silent=silent)
        else:
            c_print(f"{function_name}: failed to write {filename} in '{mode}' mode", color="RED", silent=silent)
        return exists

    else:
        c_print(f"{function_name}: failed to write {temporary_file} in '{mode}' mode", color="RED", silent=silent)
        return False


def remove_directory(directory, mode="single", silent=False):
    # setup help() return for the function
    """
    function to REMOVE DIRECTORIES

    :param directory: path to the directory the function is removing

    :param mode: function has 3 modes ("single", "all", and "clear")
                "single" mode - removes the end point target directory, ( will not remove files )

                "all" mode - removes all directories between the working directory
                             and the target directory. ( will not remove files)

                "clear" mode - removes a directory and all of its children
                               !! CAUTION YOU WILL LOSE ALL CONTENT UNDER THE TARGET DIRECTORY!!


    :param silent: a True or False bool to determine whether or not to make print statements

    :return: the function has 3 possible returns
             (True, False, None)
             True = Directory Removed
             False = Directory Removal Failed
             None = Directory Does Not Exist

    """

    # establish function name for print/debug statements
    function_name = f"{handler_name}.remove_directory()"
    silent = True if silence else silent

    # we'll just take a bit of code from the remove_file() function and turn it into
    # a reusable sub function to confirm the success of the main function, across modes
    def confirm_remove(sf_directory):
        # use the is_directory to determine whether or not we successfully deleted the directories
        # and return a bool for success/failure
        _exists = is_directory(sf_directory, silent=True)
        if _exists:
            c_print(f"{function_name}: ({os.getcwd()}\\<{sf_directory}>) unsuccessfully removed",
                    color="L_RED", silent=silent)
        # if the file still exists the function has failed remove it
        else:
            c_print(f"{function_name}: ({os.getcwd()}\\<{sf_directory}>) successfully removed",
                    color="GREEN", silent=silent)
        # return the bool exists to determine success or failure
        return _exists

    # begin function logic
    # check if the directory exists
    if is_directory(directory, silent=True):
        # and filter modes between single and all

        # "single" mode deletes the end directory only.
        if mode == "single":
            # make sure that the file directory is empty
            try:
                # use the os.rmdir() function to remove the directory
                os.rmdir(directory)

                # return whether or not the directory was deleted
                # by using the sub function
                return confirm_remove(directory)

            # if we except with an OSError, it means that the directory has contents
            except OSError:
                # if silent set to !FALSE! ( ͡ʘ ͜ʖ ͡ʘ) *slowly unwinds*
                # blah blah blah
                c_print(f"{function_name}: ({os.getcwd()}\\<{directory}>) is not empty\n"
                        f"Consider using 'clear' mode to remove the target directory, "
                        f"and all sub-directories/files",
                        color="L_RED", silent=silent)
                # and/or return as False, directory could not be removed.
                return False

        # "all" mode removes all directories between the working directory and the target directory
        elif mode == "all":
            # make sure that the file directory is empty with try/except statement
            try:
                # call os.removedirs() with the directory passed in to remove all
                # directories between working and target
                os.removedirs(directory)

                # return whether or not the directory was deleted
                # by using the sub function
                return confirm_remove(directory)

            # if we except with an OSError, it means that the directory has contents
            except OSError:
                # squaaawwwk!!
                c_print(f"{function_name}: ({os.getcwd()}\\<{directory}>) is not empty\n"
                        f"Consider using 'clear' mode to remove the target directory, "
                        f"and all sub-directories/files",
                        color="L_RED", silent=silent)
                # and/or return as False, directory could not be removed.
                return False

        # "clear" mode removes everything under a target directory
        elif mode == "clear":
            # using the shutil.rmtree(directory) function.
            shutil.rmtree(directory)

            # let's check that the directory has been deleted
            exists = is_directory(directory, silent=True)
            # FAILURE
            if exists:
                c_print(f"{function_name}: could not clear ({os.getcwd()}\\<{directory}>)",
                        color="RED", silent=silent)

            # SUCCESS
            else:
                c_print(f"{function_name}: successfully cleared ({os.getcwd()}\\<{directory}>) and "
                        f"all sub-directories/files", color="GREEN", silent=silent)
            # we return the True False bool of exists to indicate success or failure.
            return exists

    # if the directory doesn't exist
    else:
        # let the user know that the directory doesn't exist.
        c_print(f"{function_name}: could not remove ({os.getcwd()}\\<{directory}>)", color="RED", silent=silent)
        c_print(f"directory does not exist!", color="L_RED", silent=silent)
        # return that the directory is None (Does Not Exist)
        return None


# TODO: add secure delete with some sort of looped overwrite function?  Need to research before diving
#       into this one.
def remove_file(filename, silent=False):
    # setup help() return for the function
    """
    function to REMOVE FILES

    :param filename: path to the file the function is removing

    :param silent: a True or False bool to determine whether or not to make print statements

    :return: has 3 possible returns (True, False, None)
             (True, False, None)
             True = File Removed
             False = File Removal Failed
             File Does Not Exist
    """

    # establish function name for print/debug statements
    function_name = f"{handler_name}.remove_file()"
    silent = True if silence else silent
    # silently make sure the file exists
    if is_file(filename, silent=True):
        # if the file exists use the imported os module to remove it
        try:
            os.remove(filename)
        except PermissionError:
            pass

        # use the is_file to determine whether or not we successfully deleted the file
        # and return a bool for success/failure
        if not is_file(filename, silent=True):
            c_print(f"{function_name}: ({os.getcwd()}\\<{filename}>) successfully removed",
                    color="GREEN", silent=silent)

            # and/or return that we have removed the file
            return True

        # if the file still exists the function has failed remove it
        else:
            c_print(f"{function_name}: ({os.getcwd()}\\<{filename}>) unsuccessfully removed",
                    color="RED", silent=silent)

            # and/or return that we have failed to remove the file
            return False

    # if the file doesn't exist...
    else:
        # if not silent, let the user know that the file doesn't exist.
        if not silent and not silence:
            # if the file doesnt exist, then run away screaming and crying
            c_print(f"{function_name}: could not remove ({os.getcwd()}\\<{filename}>)", color="RED", silent=silent)
            c_print(f"file does not exist.", color="L_RED", silent=silent)

        # and/or return that the file is None (Does Not Exist)
        return None


def get_all_file_dirs(starting_dir, silent=False):
    """
    function to grab all file pointer directories under a target directory
    and return the output as a list of strings

    :param starting_dir: the directory under which to find all file directories
    :param silent: whether or not to print
    :return: list of file directories
    """
    """"""
    # function name
    function_name = f"{handler_name}.filter_dirs()"
    silent = True if silence else silent

    # sub function to sort to output
    def process_output(f_path, out):
        if os.path.isdir(f_path):
            out = out + get_all_file_dirs(f_path)
        else:
            out.append(f_path)
        return out

    # handle errors
    try:
        # create a list of file and sub directories
        # names in the given directory
        list_of_files = os.listdir(starting_dir)
        output = []
        # Iterate over all the entries
        for entry in list_of_files:
            # Create full path
            full_path = os.path.join(starting_dir, entry)
            # If entry is a directory then get the list of files in this directory
            # using the process_output() sub-function
            output = process_output(full_path, output)

        # return the output
        return output
    # if access denied
    except PermissionError:
        handle_errors(error="PermissionError", log=logging, silent=silent,
                      msg=f"{function_name}: {function_name}: AccessDenied")
    # if file not found
    except FileNotFoundError:
        handle_errors(error="FileNotFoundError", log=logging, silent=silent,
                      msg=f"{function_name}: {function_name}: File Not Found!")


def filter_dirs(dir_list, extension=None):
    """
    function to return an optionally filtered list of file pointer directories
    returned by the get_all_file_dirs() function

    :param dir_list: a list of directories to filter and pass
    :param extension: a .abc extension filter, only returns files with specified extension
    :return: returns a list comprehension of (optionally extension filtered) file directories
    """
    # function name
    function_name = f"{handler_name}.filter_dirs()"
    # handle potential errors
    try:
        # if the extension argument is is unfilled
        if extension is None:
            # we'll just return the dir_list
            return dir_list
        # otherwise
        else:
            # we will return a list comprehension that only contains file pointers that
            # contain the .extension argument string
            return [directory for directory in dir_list if extension in directory]
    #
    # type error
    except TypeError:
        return f"{function_name}: TypeError"


def print_dirs(dir_list, mode="normal", silent=False):
    """
    function that prints a line spaced directories -- or can (optionally) print them
    as a formatted 'list'-alike set of strings that can be copied and pasted into code

    :param dir_list: list of strings containing directories
    :param mode: (normal, formatted)
                    normal - print line spaced and stripped of container clutter
                    formatted - made to look like a python code formatted
                                line separated list
    :param silent: whether or not to print
"""
    # do a before loop conditional
    # to check the mode
    if mode == "formatted":
        # and make a decorative print to kick off the output before the loop begins
        c_print("f_dir_list = ")

    # enumerate the dir_list to get the index and file pointer strings
    for index, system_link in enumerate(dir_list):
        # make sure the item in the list is a string otherwise pass around for the next loop
        if isinstance(system_link, str):
            if mode == "normal":
                c_print(system_link)
            elif mode == "formatted":
                # slice the system_link so that the output is usable in code
                s_link = system_link.replace("/", "\\")
                s_link = s_link.replace("\\", "\\\\")

                # handle index positions to craft a list like console output using the print statement
                # handling the first line
                if index == 0:
                    c_print(f"['{s_link}',", silent=silent)

                # all in-between lines
                elif index + 1 < len(dir_list):
                    c_print(f" '{s_link}',", silent=silent)

                # and the last line
                else:
                    c_print(f" '{s_link}']", silent=silent)


def copy_file(src_file, dest_file=None, mode="copy", overwrite=False, silent=False):
    """
    function to copy individual files into a new directory with an optionally formatted name

    :param src_file: file to be copied
    :param dest_file: destination of copied file
    :param mode: shutils -- (copy, copy2, copyfile)
    :param overwrite: whether or not to delete the original file and try again when confronted with a FileExistsError
    :param silent: whether or not to print debug statements
    :return: 
    """
    # function name
    function_name = f"{handler_name}.copy_file()"
    silent = True if silence else silent
    # possible mode calls
    copy_mode = (1, "c", "C", "copy", "Copy", "COPY")
    copy2_mode = (2, "c2", "C2", "copy2", "Copy2", "COPY2")
    copyfile_mode = (3, "cf", "Cf", "CF", "copyfile", "Copyfile", "CopyFile", "COPYFILE")

    # sub function to do all the work
    def c_file(s_file, d_file, _mode):

        failure = False
        if _mode in copy_mode:
            try:
                # try to copy the file
                shutil.copy(s_file, d_file, follow_symlinks=True)
            except FileNotFoundError:
                failure = handle_errors("FileNotFoundError", silent=silent)
            except PermissionError:
                failure = handle_errors("PermissionError", silent=silent)
            except FileExistsError:
                if not overwrite:
                    failure = handle_errors("FileExistsError", silent=silent)
                # if file exists and is set to overwrite remove the file in the way
                # and try again
                else:
                    remove_file(dest_file)
                    c_file(src_file, dest_file, _mode)
            except OSError:
                failure = handle_errors("OSError", silent=silent)

        elif mode in copy2_mode:
            try:
                shutil.copy2(src_file, dest_file, follow_symlinks=True)
            except FileNotFoundError:
                failure = handle_errors("FileNotFoundError", silent=silent)
            except PermissionError:
                failure = handle_errors("PermissionError", silent=silent)
            except FileExistsError:
                if not overwrite:
                    failure = handle_errors("FileExistsError", silent=silent)
                # if file exists and is set to overwrite remove the file in the way
                # and try again
                else:
                    remove_file(dest_file)
                    c_file(src_file, dest_file, _mode)
            except OSError:
                failure = handle_errors("OSError", silent=silent)

        elif mode in copyfile_mode:
            try:
                shutil.copyfile(src_file, dest_file, follow_symlinks=True)
            except FileNotFoundError:
                failure = handle_errors("FileNotFoundError", silent=silent)
            except PermissionError:
                failure = handle_errors("PermissionError", silent=silent)
            except FileExistsError:
                if not overwrite:
                    failure = handle_errors("FileExistsError", silent=silent)
                # if file exists and is set to overwrite remove the file in the way
                # and try again
                else:
                    remove_file(dest_file)
                    c_file(src_file, dest_file, _mode)
            except OSError:
                failure = handle_errors("OSError", silent=silent)

        else:
            failure = handle_errors("ModeError", silent=silent)

        if is_file(dest_file, silent=True) and not failure and file_verification(src_file, dest_file, silent=True):
            c_print(f"{function_name}: successfully copied\nfrom:({src_file})\nto: ({dest_file})",
                    color="GREEN", silent=silent)
        else:
            c_print(f"{function_name}: something went wrong while copying:\nfrom:({src_file})\nto: ({dest_file})",
                    color="RED", silent=silent)

    if dest_file is None:
        dest_file = filedialog.askdirectory()
        dest_file = uniform_filename(dest_file)

    # run sub-function
    c_file(src_file, dest_file, mode)


def mass_copy(file_list, dest_dir=None, mode="c", silent=False):
    """
    function to copy an entire list of files into a single directory

    :param file_list: list or tuple containing file pointing directories
    :param dest_dir: destination directory to copy the list of files
    :param mode: to copy in c, c2, cfile (shutil.copy/copy2/copyfile)
    :param silent: whether or not to print statements
    """
    function_name = f"{handler_name}.mass_copy()"
    silent = True if silence else silent

    copy_mode = (1, "c", "C", "copy", "Copy", "COPY")
    copy2_mode = (2, "c2", "C2", "copy2", "Copy2", "COPY2")
    copyfile_mode = (3, "cf", "Cf", "CF", "copyfile", "Copyfile", "CopyFile", "COPYFILE")

    if dest_dir is None:
        dest_dir = filedialog.askdirectory()
        dest_dir = uniform_filename(dest_dir)

    for index, system_link in enumerate(file_list):
        if mode in copy_mode or mode in copy2_mode or mode in copyfile_mode:
            # c_print(f"copying {system_link} to {dest_dir} in {mode}")
            try:
                copy_file(system_link, dest_dir, mode)
            except ValueError:
                handle_errors(error="ValueError", log=logging, silent=silent)
            except TypeError:
                handle_errors(error="TypeError", log=logging, silent=silent)

        else:
            c_print(f"{function_name}: copy mode out of range; defaulting to copy mode", color="L_RED", silent=silent)
            copy_file(system_link, dest_dir, "c", silent=silent)


def download_file(file_url, extension=None, directory=None, silent=False):
    """

    :param file_url:
    :param extension:
    :param directory:
    :param silent:
    :return:
    """

    function_name = f"{handler_name}.download_file()"
    silent = True if silence else silent
    failure = False

    # def bar_custom(current, total, width=80):
    #     if not silent:
    #         sl_print(f"Downloading: {current / total * 100:.2f} [{current} / {total}] bytes")

    if directory == "gui":
        file_name = filedialog.asksaveasfilename()
        file_name = uniform_filename(file_name)
    else:
        # set up directory if applicable
        if directory:
            directory = uniform_filename(directory)

        # clean up url name to get file name
        file_name = file_url.split('/')[-1]
        # append .file extension to filename
        if "." not in file_name and extension:
            extension_separator = "." if "." not in extension else ""
            file_name = f'{file_name}{extension_separator}{extension}'

        # append directory if applicable
        if directory is not None:
            file_name = f"{directory}\\{file_name}" \
                if directory[-1] != "\\" and directory[-1] != "\\\\" and directory[-1] != "/" \
                else f"{directory}{file_name}"

    if not silent:
        sl_print(f"{function_name}: downloading {file_url} to {file_name}")
    # download file
    wget.download(file_url, out=file_name)

    if is_file(file_name, silent=True) and not failure:
        if not silent:
            sl_print(f"{function_name}: successfully downloaded {file_url} to {file_name}\n")
    else:
        if not silent:
            sl_print(f"{function_name}: unsuccessfully downloaded {file_url} to {file_name}\n")


if __name__ == '__main__':
    test_file = "https://freesound.org/data/previews/12/12691_32572-lq.mp3"
    download_file(test_file, directory="gui")



import datetime
import os
import sys

verbose = 0
initialized = False

try:
    # import colorama
    import colorama
except ImportError:
    print("PrintHandler -- could not import colorama -- color function will be disabled")
    colorama = None


def init(verbosity=0):
    """
    Initializes colorama and adjusts log verbosity level

    :param verbosity: control the print level when log_mode is set
                      0 - Print all debug messages - *Default
                      1 - Print on Info -> Error only
                      2 - Print on Debug -> Error only
                      3 - Print on Warning -> Error only
                      4 - Print on Error only
    """
    # initialize colorama
    if not initialized:
        global verbose
        verbose = verbosity
        colorama.init()
        print("PrintHandler initialized")


def sl_print(msg):
    sys.stdout.write(f"\r{msg}")
    sys.stdout.flush()


def write_log(message=None, logfile=None):
    # setup help() return for the function
    """
    function to write messages to a logfile

    :param message: message you want to log
    :param logfile: path to the file you want to act as your logfile
    :return: has no return
    """

    # get the current time as of function call
    now = str(datetime.datetime.now())[:-7]

    # if there is no message skip the nonsense
    if message is None:
        pass
    # if there is a message to write, continue
    else:
        # if the logfile parameter is empty, default to "logfile.txt" otherwise keep parameter entry
        logfile = "logfile.txt" if logfile is None else logfile
        # add a new line to the end of the write if it isn't already present in the content
        content = f"({now})  - {message}\n" if message[-1] != "\n" else f"({now})  - {message}"
        # # set the mode to append if the file already exists or write a new file otherwise
        mode = "a" if os.path.isfile(logfile) else "w"
        # # write log file
        with open(logfile, mode) as f:
            f.write(content)


# function to act as a custom print feature
def c_print(message=None, sep=None, end=None, file=None, flush=False,
            color=None, timestamp=False, log_mode=None, silent=False):
    # setup help() return for the function
    """
    function to act as a custom print feature

    :param message: the string that you want to push to console

    :param sep: separator in print()
    :param end: end of string in print()
    :param file: file in print()
    :param flush: flush in print()

    :param color: color setting of print
                  BLACK, RED, L_RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE

    :param timestamp: whether or not to print with a timestamp

    :param log_mode: what warning level to print in if any
                     0 - None
                     1 - INFO - CYAN
                     2 - DEBUG - YELLOW
                     3 - WARNING - L_RED
                     4 - ERROR - RED

    :param silent: bool, True, print or False, do not print
    :return: has no return
    """

    # verbosity level of print (log_mode)
    log_modes = (0, 1, 2, 3, 4)
    if log_mode in log_modes:
        if log_mode == 0:
            log_mode = None

        # INFO
        if log_mode == 1:
            color = "CYAN"
            message = f"INFO: {message}"

        # DEBUG
        if log_mode == 2:
            color = "YELLOW"
            message = f"DEBUG: {message}"

        # WARNING
        if log_mode == 3:
            color = "L_RED"
            message = f"WARNING: {message}"

        # ERROR
        if log_mode == 4:
            color = "RED"
            message = f"ERROR: {message}"

    # if our log_mode is out of range, just set it to None
    else:
        log_mode = None

    # if our timestamp is set to True add a datetime.datetime.now stamp to our message
    if timestamp:
        # datetime.datetime.now() is not subscriptable, so we'll convert it into a string
        # to where we can slice it with indexing to erase the unwanted tail at the end
        now = str(datetime.datetime.now())
        message = f"({now[:-7]}) - {message}"

    # do not print log statements lower than the settings.py allows and the log_mode is acceptable
    do_not_print = True if log_mode in log_modes and verbose > log_mode else False

    # check if we need to skip the print and pass out of the function
    if do_not_print:
        pass

    # otherwise we continue
    else:
        # list of acceptable colors
        colors = ("BLACK", "RED", "L_RED", "GREEN", "YELLOW", "BLUE", "MAGENTA", "CYAN", "WHITE")
        # if the color isn't in the list of colors or is otherwise empty
        if color is None or color not in colors:
            # make sure we print a freshly reset message as to remove any previous formatting from
            # previous calls
            if colorama is None and not silent:
                print(f"{message}",
                      sep=sep, end=end, file=file, flush=flush)
            elif colorama and not silent:
                if not silent:
                    print(f"{colorama.Fore.RESET}{message}",
                          sep=sep, end=end, file=file, flush=flush)

        # if our log_mode is None or it is acceptable, we assign colors to the message
        if log_mode is None or log_mode in log_modes:
            # write log_mode messages to a log file
            if log_mode in log_modes and log_mode is not None:
                if log_mode >= verbose:
                    write_log(message)

            if colorama is None:
                if not silent:
                    print(f"{message}",
                          sep=sep, end=end, file=file, flush=flush)
            else:
                if not silent:
                    # add color to and print the message, and close
                    # with a Fore.RESET to prevent color settings from bleeding
                    # into subsequently printed statements
                    if color == "BLACK":
                        print(f"{colorama.Fore.WHITE}{message}{colorama.Fore.RESET}",
                              sep=sep, end=end, file=file, flush=flush)
                    if color == "L_RED":
                        print(f"{colorama.Fore.LIGHTRED_EX}{message}{colorama.Fore.RESET}",
                              sep=sep, end=end, file=file, flush=flush)
                    if color == "RED":
                        print(f"{colorama.Fore.RED}{message}{colorama.Fore.RESET}",
                              sep=sep, end=end, file=file, flush=flush)
                    if color == "GREEN":
                        print(f"{colorama.Fore.GREEN}{message}{colorama.Fore.RESET}",
                              sep=sep, end=end, file=file, flush=flush)
                    if color == "YELLOW":
                        print(f"{colorama.Fore.YELLOW}{message}{colorama.Fore.RESET}",
                              sep=sep, end=end, file=file, flush=flush)
                    if color == "BLUE":
                        print(f"{colorama.Fore.BLUE}{message}{colorama.Fore.RESET}",
                              sep=sep, end=end, file=file, flush=flush)
                    if color == "MAGENTA":
                        print(f"{colorama.Fore.MAGENTA}{message}{colorama.Fore.RESET}",
                              sep=sep, end=end, file=file, flush=flush)
                    if color == "CYAN":
                        print(f"{colorama.Fore.CYAN}{message}{colorama.Fore.RESET}",
                              sep=sep, end=end, file=file, flush=flush)
                    if color == "WHITE":
                        print(f"{colorama.Fore.BLACK}{message}{colorama.Fore.RESET}",
                              sep=sep, end=end, file=file, flush=flush)



def is_float(number):
    """
    function to determine if the input value can be converted into a float
    by using the try/except technique, filtering out True and 'NaN' (which can be converted to floats)

    :param number: the value to test
    :return: True/False
             if True, the value can be converted to a float
             if False, the value cannot be converted to a float
    """
    # return False if input is NaN as it can be converted to a float.
    try:
        n = str(number)
    except ValueError:
        return False
    except TypeError:
        return False

    false_floats = (str(True), str(False), str(None), "Nan")
    if n in false_floats:
        return False

    # try/except block to attempt converting the input into a float and then  returning True
    # or return False if there is an exception
    try:
        float(number)
        return True
    except ValueError:
        return False
    except TypeError:
        return False


def safe_sum(*numbers):
    """
    function to safely get the sum value of an input
    at its core it can average ints, floats, and strings while ignoring bad data types
    it can also process recursion loops of nested lists and tuples

    this is a safer and more capable substitute for python's built in sum() function

    :param numbers: the numbers to sum
    :return: returns the sum of the numbers argument
    """
    # name function for error reporting
    function_name = f"safe_sum()"

    # reusable subfunction to do all the work
    def sum_recursion(nums):
        """
        a helper subfunction used by safe_sum to do all of the lifting including counting, number string handling,
        recursion loops through nested loops and tuples by calling itself
        when hitting tuples or lists inside of its input, and filtering bad value/data types

        :param nums: number data to sum, seperated by comma
        :return: returns the sum of nums input
        """
        # name subfunction for error reporting
        subfunction_name = f"{function_name}.sum_recursion()"

        # check that the nums variable is not empty and begin work
        if nums is not None:
            # set counter
            s_of_nums = 0

            # start for loop to process the subfunction input
            for num in nums:
                # check if num in nums is an int or a float and count
                if isinstance(num, int) or isinstance(num, float):
                    if is_float(num):
                        s_of_nums += num

                # check if num in nums is a string and that it can be converted to a float
                # using is_float(num) and converting it to a float if it has a '.' present or an int otherwise
                # and then adding it to the sum
                elif isinstance(num, str):
                    if is_float(num):
                        if "." in num:
                            s_of_nums += float(num)
                        else:
                            s_of_nums += int(num)

                # check if the num in nums is a tuple or list, and that the list is not empty
                elif isinstance(num, tuple) or isinstance(num, list):
                    # try/except block to catch recursion errors and pass if the subfunction excepts
                    try:
                        # attempt to add the output of a recursion loop
                        s_of_nums += sum_recursion(num)

                    # if the subfunction hits the python recursion limit announce and return 0
                    except RecursionError:
                        print(f"{subfunction_name}: Recursion limit reached during nested tuple/list loop!")

            # return the final sum output of the subfunction
            return s_of_nums

        # if input is None return 0
        else:
            return 0

    # begin function code
    # make sure the numbers variable isn't empty
    if len(numbers) > 0:
        # catch potential recursion errors for extreme depth nested lists or tuples
        # call the recursive summing subfunction and place its return into numbers_sum
        # and return the sum of the input numbers
        try:
            numbers_sum = sum_recursion(numbers)
            return numbers_sum

        # if the function hits the python recursion limit announce and return 0
        except RecursionError:
            print(f"{function_name}: recursion limit reached during nested recursion loop!")
            return 0

    # if the length of the numbers(*args) is empty
    # return a sum of 0
    else:
        return 0


def safe_average(*numbers):
    """
    function used to average an input safely by filtering bad value/data types
    and handling int, float, number string, list, and tuples
    including the ability to do recursion loops through nested lists/tuples
    when hitting additional lists/tuples inside of lists/tuples.

    :param numbers: the numbers to get the average / mean of
    :return: returns the average / mean of the numbers input
    """
    # name function for error reporting
    function_name = f"safe_average()"

    # create a sub function to handle the bulk of the work
    def avg_recursion(nums):
        # set up sub-function help string
        """
        a helper subfunction used by safe_average to do all of the lifting and counting
        and even do recursion loops through nested loops and tuples by calling itself when
        hitting tuples or lists inside of its input.
        
        :param nums: the tuple/list to do a recursion loop through
        :return: returns the sum of numbers and total number of numbers on index [0] and [1]
                 respectfully
        """
        # name subfunction for error reporting
        subfunction_name = f"{function_name}.safe_recursion()"
        # check that the subfunction input is not empty and begin work
        if len(nums) > 0:
            # set up the counter variables
            s_of_nums = 0
            l_of_nums = 0
            
            # start for loop to cycle through contents of nums and build the sum and length
            for num in nums:
                # check if num is an int or a float and count if so
                if isinstance(num, int) or isinstance(num, float):
                    if is_float(num):
                        s_of_nums += num
                        l_of_nums += 1

                # check if num in nums is a string and that it can be converted to a float
                # using is_float(num) and converting it to a float if it has a '.' present or an int otherwise
                # and then adding the sum and l_of_nums
                elif isinstance(num, str):
                    if is_float(num):
                        if '.' in num:
                            s_of_nums += float(num)
                        else:
                            s_of_nums += int(num)
                        l_of_nums += 1

                # check if num is a nested tuple or list
                elif isinstance(num, tuple) or isinstance(num, list):
                    # catch block to catch recursion errors and pass if the subfunction excepts
                    try:
                        # and get the sum and length returns by doing a recursion loop through the tuple/list
                        # using the safe_recursion subfunction
                        _n1, _n2 = avg_recursion(num)
                        s_of_nums += _n1
                        l_of_nums += _n2

                    # if the subfunction hit the python recursion limit announce and return 0
                    except RecursionError:
                        print(f"{subfunction_name}: Recursion limit reached during nested tuple/list loop!")
            
            # return the sums and the lengths of all the numbers
            return s_of_nums, l_of_nums

    # begin function code
    # make sure the numbers variable isn't empty
    if len(numbers) > 0:
        # catch potential recursion errors for extreme depth nested lists or tuples
        # call the recursive averaging subfunction and place index 0 and 1 into numbers_sum and numbers_length
        # and return the average of the input by dividing sum by the number of numbers
        try:
            numbers_sum, numbers_length = avg_recursion(numbers)
            return numbers_sum / numbers_length

        # if the function hit the python recursion limit announce and return 0
        except RecursionError:
            print(f"{function_name}: recursion limit reached during nested recursion loop!")
            return 0
    
    # if the length of the numbers(*args) is empty
    # return an average of 0
    else:
        return 0


def get_horizon(vantage_height, h_constant=1.5):
    """
    gets the viewing distance to the horizon on earth by height of the eyes

    :param vantage_height: in feet
    :param h_constant: horizon constant :shrug:
    :return: returns horizon in miles
    """
    return (h_constant * vantage_height) ** 0.5


def emc2(mass):
    """
    function that gets the energy equivalence of a given mass in kilograms using Einstein's e=mc2 equation
    and returns the energy equivalence (e) in joules

    :param mass: mass is the mass of the object in kilograms
    :return: returns e, as mass energy equivalence measured in joules
    """
    # return mass times c squared
    return mass * (299_792_458 ** 2)


# TESTING
if __name__ == '__main__':
    help(is_float)
    help(safe_sum)
    help(safe_average)

    # sample should return 2.5 / 10 respectfully
    calibration_sample_clean = [1, 2, 3, 4]
    calibration_sample_easy = [1, [2, (3, 4)]]
    calibration_sample_hard = [1, "two", True, ["2", ("3.0", 4)]]
    # impossible? sum: 1231 and avg: 49.24"
    crash_test_sample = (1, 2, "lol", [1, 2, "buckle my shoe", 3, 4, 5.0, "6", "7.0"],
                         (10, 20, [30, 40, (1, 2, 3, 4, 5, 1000, ("lol", "25", 50))]), 1, 2, 3, 4)

    print("TESTING MathHandler")
    print("Calibration tests, should return a sums of 10 and averages of 2.5")
    print(f"Clean: {calibration_sample_clean}")
    print(f"Easy: {calibration_sample_easy}")
    print(f"Hard: {calibration_sample_hard}")
    print("'Impossible' test should return sums of 1231 and averages of 49.24")
    print(f"Impossible: {crash_test_sample}")

    print("")

    print("TESTING safe_sum() and safe_average() FUNCTIONS")
    print("calibration test (easy mode), should return sum: 10 and avg: 2.5")
    print(f"sum: {safe_sum(calibration_sample_easy)}")
    print(f"avg: {safe_average(calibration_sample_easy)}")

    print("")

    print("calibration test (hard mode), should return sum: 10 and avg: 2.5")
    print(f"sum: {safe_sum(calibration_sample_hard)}")
    print(f"avg: {safe_average(calibration_sample_hard)}")

    print("")

    print("impossible test: should return sum: 1231 and avg: 49.24")
    try:
        print(f"sum: {safe_sum(crash_test_sample)}")
        print(f"avg: {safe_average(crash_test_sample)}")
    except TypeError:
        print("Crash Test Failed! TypeError")
    except ValueError:
        print("Crash Test Failed! TypeError")
    except ZeroDivisionError:
        print("Crash Test Failed! ZeroDivisionError")

    print("")
    print("TEST BUILT-IN PYTHON 'sum()' FUNCTION")
    print("Python's native sum() function can handle clean passes, but its robustness ends there.")
    print("The sum() function is very fast, but it cannot process dirty/convoluted data, and it cannot")
    print("recurse through nested lists or tuples")
    print("")
    try:
        print(f"clean test: {sum(calibration_sample_clean)}")
    except TypeError:
        print("Clean Test Failed! TypeError")
    except ValueError:
        print("Clean Test Failed! TypeError")

    try:
        print(f"easy test: {sum(calibration_sample_easy)}")
    except TypeError:
        print("Easy Test Failed! TypeError")
    except ValueError:
        print("Easy Test Failure: TypeError")

    try:
        print(f"hard test: {sum(calibration_sample_hard)}")
    except TypeError:
        print("Hard Test Failed! TypeError")
    except ValueError:
        print("Hard Test Failed! TypeError")

    try:
        print(f"impossible test: {sum(crash_test_sample)}")
    except TypeError:
        print("Impossible Test Failed! TypeError")
    except ValueError:
        print("Impossible Test Failed! TypeError")

    print("")
    print("")

    print("e=mc2")
    help(emc2)
    print("testing ecm2()")
    print(f"1Kg of mas is equivalent to {emc2(1)/1000/1000/1000/1000/1000} Petajoules of energy.")



#
# If you're reading this...
# Hi!  I'm Billy B. Bobby Bobertson Jr. III. -- my friend's call me Bob...
# I hope you are well, and are enjoying yourself wherever you may be. <3
#
# And If you are not reading this... ^^ this ^^ to you ALL the same (@everyone).
#



import os
import re

import setuptools


# thanks colorama
def read_file(path):
    with open(os.path.join(os.path.dirname(__file__), path)) as f:
        return f.read()


# thanks colorama
def _get_version_match(content):
    # Search for lines of the form: # __version__ = 'ver'
    regex = r"^__version__ = ['\"]([^'\"]*)['\"]"
    version_match = re.search(regex, content, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


# thanks colorama
def get_version(path):
    return _get_version_match(read_file(path))


NAME = "Junkyard"
# thanks colorama
setuptools.setup(
    name=NAME,
    version=get_version(os.path.join('junkyard', '__init__.py')),
    author="CKD",
    author_email="crushkilldestroy@tuta.io",
    description="A collection of [certifiable garbage] dev helpers/handlers created by and for the author",
    long_description=read_file('README.md'),
    long_description_content_type="text/markdown",
    keywords='personal library, not for you, unless you want it',
    url="https://github.com/crushkilldestroy/Junkyard",
    packages=[NAME],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Software Development",
    ],
    install_requires=[
        'colorama',
        'wget',
    ],
    python_requires='>=3.7',
)



# import standard library modules
import datetime
import os
# import random
import shutil
# import stat
import sys
import time
# import tempfile


"""
JUNKYARD IS A COMPILATION OF TID BIT (probably actual CERTIFIED *garbage*) HELPERS 
CRAFTED DURING PROJECTS TO BE REUSED AT A LATER TIME BY THE AUTHOR.

THIS IS THE HANDLER FILE, OTHERWISE KNOWN AS THE GARBAGE BASKET 
THIS IS WHERE THE DIFFERENT CLASSES AND METHODS/FUNCTIONS CAN BE FOUND 
WITHIN THE MODULE
"""


# DIAGNOSTICHANDLER
# CRYPTOHANDLER
# FILEHANDLER
# PRINTHANDLER


# DIAGNOSTICS HANDLER
class DiagnosticHandler:
    #
    def __init__(self):
        # setup help() return for the class
        """
        WORK IN PROGRESS

        DiagnosticHandler Class Initializer method.
        establishes various class wide variables
        """
        self.class_name = "DiagnosticHandler()"
        print(f"{self.class_name} Initialized")

    def speed_test(self, subject=None, n_tests=100, float_precision=8, loop_time=0.1, *args):
        # setup help() return for the method
        """
        method for testing the speed of a function
        :param subject: function to test without ()'s
        :param n_tests: number of tests to run and average
        :param float_precision: floating point precision of the timer
        :param loop_time: break time between each test (in seconds)
        :param args: parameters for subject_1 and subject_2
        :return: returns nothing, just a way to break out of the function if subject is empty
        """

        # establish method name for print/debug statements
        method_name = f"{self.class_name}.speed_test()"

        # subject = # None
        # n_tests = # 100
        # float_precision = # 8
        # loop_time = # 0.1

        # make sure subjects have values
        if subject is None:
            print(f"{method_name}: subject is empty")
            return

        # SUBJECT TEST LOOP
        print("TESTING SUBJECT 1")
        subject_times = []
        # start loop for n_tests times
        for x in range(n_tests):
            # start timer
            timer_start = time.perf_counter()

            # try to handle potential errors
            try:
                # call our first test subject from arguments
                subject(*args)
            except ValueError:
                print(f"{method_name}: ({subject}({args})) failed to run properly")
                break
            except TypeError:
                print(f"{method_name}: ({subject}({args})) failed to run properly")
                break
            except FileNotFoundError:
                print(f"{method_name}: ({subject}({args})) failed to run properly")
                break
            except OSError:
                print(f"{method_name}: ({subject}({args})) failed to run properly")
                break

            # stop timer
            timer_stop = time.perf_counter()

            # get total time
            subject_time = timer_stop - timer_start
            print(f"[Test {x+1}] {subject_time} seconds")
            # and append it to our times list
            subject_times.append(subject_time)
            # take a break
            time.sleep(loop_time)

        # get averages of tests
        subject_average = round(sum(subject_times) / len(subject_times), float_precision)

        # create a line space
        print("")

        # print Subject wins and average time
        print(f"Subject: average runtime of {subject_average} seconds")

    def speed_comparison_test(self, subject_1=None, subject_2=None,
                              n_tests=100, float_precision=8, intermission_time=1, loop_time=0.1, *args):
        # setup help() return for the method
        """
        method for testing the speed difference of two functions

        :param subject_1: first function to test without ()'s
        :param subject_2: first function to test without ()'s
        :param n_tests: number of tests to run and average
        :param float_precision: floating point precision of the timer
        :param intermission_time: break time between the two sets of tests (in seconds)
        :param loop_time: break time between each test (in seconds)
        :param args: parameters for subject_1 and subject_2
        :return: returns nothing, just a way to break out of the function if subject 1 or 2 are empty
        """

        # establish method name for print/debug statements
        method_name = f"{self.class_name}.speed_comparison_test()"

        # make sure subjects have values
        if subject_1 is None:
            print("subject_1 is empty")
            return
        if subject_2 is None:
            print("subject_2 is empty")
            return

        # SUBJECT 1 TEST LOOP
        print("TESTING SUBJECT 1")
        subject_1_times = []
        # start loop for n_tests times
        for x in range(n_tests):
            # start timer
            timer_start = time.perf_counter()

            # try to handle potential errors
            s1_failure_msg = f"{method_name}: ({subject_1}({args})) failed to run properly"
            try:
                # call our first test subject from arguments
                subject_1(*args)
            except ValueError:
                print(s1_failure_msg)
                break
            except TypeError:
                print(s1_failure_msg)
                break
            except FileNotFoundError:
                print(s1_failure_msg)
                break
            except OSError:
                print(s1_failure_msg)
                break

            # stop timer
            timer_stop = time.perf_counter()

            # get total time
            subject_1_time = timer_stop - timer_start
            print(f"[Test {x + 1}] {subject_1_time} seconds")
            # and append it to our times list
            subject_1_times.append(subject_1_time)
            # take a break
            time.sleep(loop_time)

        # TAKE A BREAK!
        print("")
        print("-- intermission --")
        print("")
        time.sleep(intermission_time)

        # SUBJECT 2 TEST LOOP
        print("TESTING SUBJECT 2")
        subject_2_times = []
        for x in range(n_tests):
            # start timer
            timer_start = time.perf_counter()

            # try to handle potential errors
            s2_failure_msg = f"{method_name}: ({subject_2}({args})) failed to run properly"
            try:
                # call our first test subject from arguments
                subject_2(*args)
            except ValueError:
                print(s2_failure_msg)
                break
            except TypeError:
                print(s2_failure_msg)
                break
            except FileNotFoundError:
                print(s2_failure_msg)
                break
            except OSError:
                print(s2_failure_msg)
                break

            # stop timer
            timer_stop = time.perf_counter()

            # get total time
            subject_2_time = timer_stop - timer_start
            print(f"[Test {x + 1}] {subject_2_time} seconds")
            # and append it to our times list
            subject_2_times.append(subject_2_time)
            # take a break
            time.sleep(loop_time)

        # get averages of tests
        subject_1_average = round(sum(subject_1_times) / len(subject_1_times), float_precision)
        subject_2_average = round(sum(subject_2_times) / len(subject_2_times), float_precision)

        # get win times for each subject using list comprehensions
        # subject 1
        subject_1_wins = [subject_1_times[index] for index, e in enumerate(subject_1_times)
                          if subject_1_times[index] < subject_2_times[index]]
        #
        # subject 2
        subject_2_wins = [subject_2_times[index] for index, e in enumerate(subject_2_times)
                          if subject_1_times[index] > subject_2_times[index]]

        # get number of wins
        subject_1_n_wins = len(subject_1_wins)
        subject_2_n_wins = len(subject_2_wins)

        # create a line space
        print("")

        # print Subject wins and average time
        print(f"Subject 1 Wins: {subject_1_n_wins} ({round((subject_1_n_wins / n_tests) * 100, 2)}%) with an "
              f"average time of {subject_1_average} seconds")
        print(f"Subject 2 Wins: {subject_2_n_wins} ({round((subject_2_n_wins / n_tests) * 100, 2)}%) with an"
              f" average time of {subject_2_average} seconds")

        # create a line space
        print("")

        # compare subject number of wins
        if subject_1_n_wins > subject_2_n_wins:
            print(f"SUBJECT 1 HAS THE MOST WINS ({subject_1_n_wins} / {n_tests})")
        elif subject_1_n_wins < subject_2_n_wins:
            print(f"SUBJECT 2 HAS THE MOST WINS ({subject_2_n_wins} / {n_tests})")
        elif subject_1_n_wins == subject_2_n_wins:
            print("SUBJECT 1 AND SUBJECT 2 HAVE THE SAME NUMBER OF WINS - TIE!")

        # compare subject average times
        if subject_1_average < subject_2_average:
            print(f"SUBJECT 1 HAD THE FASTEST AVERAGE @ {subject_1_average} SECONDS")
            print(f"or {subject_2_average / subject_1_average} times Subject 2's average")
        elif subject_1_average > subject_2_average:
            print(f"SUBJECT 2 HAD THE FASTEST AVERAGE @ {subject_2_average} SECONDS")
            print(f"or {subject_1_average / subject_2_average} times Subject 1's average")
        elif subject_1_average == subject_2_average:
            print(f"SUBJECT 1 AND SUBJECT 2 HAVE THE SAME AVERAGE @ {subject_1_average} SECONDS")


# ON HOLD
# class to manage encryption and decryption of bytes files
# TODO: ENCRYPTION!
class CryptoHandler:
    # class initializer
    def __init__(self):
        self.class_name = "CryptoHandler()"
        try:
            # try to import our cryptography modules
            import base64
            import logging

            from cryptography.exceptions import UnsupportedAlgorithm
            from cryptography.hazmat.backends import default_backend
            from cryptography.hazmat.primitives import hashes
            from cryptography.hazmat.primitives.asymmetric import padding
            from cryptography.hazmat.primitives.asymmetric import rsa

            # push our modules into class wide variables
            # other modules
            self.base64 = base64
            self.logging = logging
            # cryptography module
            self.UnsupportedAlgorithm = UnsupportedAlgorithm
            self.default_backend = default_backend
            self.hashes = hashes
            self.padding = padding
            self.rsa = rsa

            # declare other class wide variables
            # declare the mode of our AES
            # self.aes_mode = self.AES.MODE_CBC
            # this should stay as 64 * 1024
            # self.chunk_size = 64 * 1024
            # this should stay as 16
            # self.str_length = 16

            print("CryptoHandler() Initialized")

        except ImportError:
            print(f"{self.class_name} Initialization Failed!")
            print(f"Could not import cryptography modules")

    # method to encrypt files
    def encrypt(self, key, filename):
        method_name = f"{self.class_name}.encrypt()"
        pass

    # method to decrypt files
    def decrypt(self, key, filename):
        method_name = f"{self.class_name}.decrypt()"

    # module to hash passwords
    def get_key(self, password):
        method_name = f"{self.class_name}.get_key()"
        pass


# WORK IN PROGRESS
# class to manage basic file system tasks in a safe and reliable way.
#
# TODO: (?)NEXT COMMIT(?) - EXPAND FUNCTIONALITY OF FILEHANDLER?
#                               - CHMOD (ON HOLD)
#                               - CHDIR / CD (Change Directory) (ON HOLD)
#                               - COPY (ON HOLD)
#                               - MOVE (ON HOLD)
#                               - FIND(?) (ON HOLD)
class FileHandler:
    def __init__(self, silent=False):
        # setup help() return for the class
        """
        WORK IN PROGRESS

        FileHandler Class Initializer method.
        establishes various class wide variables

        *all methods in class can be silenced with silent=True*
        * BUT OTHERWISE HAVE INFORMATIVE PRINT STATEMENTS *

        :param silent: a True or False bool to determine whether or not to make print statements
        """
        # establish class name for print/debug statements
        self.class_name = "FileHandler()"
        # if silent is false
        if not silent:
            # do a sys.stdout.write and flush at top and bottom to print on same line
            sys.stdout.write(f"\r{self.class_name} Initializing")
            sys.stdout.flush()
        
        # initialize all of the class wide variables
        # standard read/write/append modes
        self.write_modes = ("w", "write", "Write", "WRITE")
        self.write_append_modes = ("a", "append", "Append", "APPEND")
        self.read_modes = ("r", "read", "Read", "READ")
        # TODO: do we need this?
        #       do we want to add the r+ / rb+ functionality, or do our current
        #       tools suffice?
        self.read_write_modes = ("r+", "R+", "rw", "RW", "readwrite", "READWRITE")

        # bytes modes
        self.write_bytes_modes = ("wb", "WB", "w_bytes", "W_bytes")
        self.write_append_bytes_modes = ("ab", "AB", "a_bytes", "A_bytes")
        self.read_bytes_modes = ("rb", "RB", "r_bytes", "R_bytes")
        # TODO: do we need this?
        #       do we want to add the r+ / rb+ / r+b functionality, or do our current
        #       tools suffice?
        self.read_write_bytes_modes = ("rb+", "RB+", "rw_bytes", "RW_bytes")
        # TODO: What about other modes like "x"?

        # CHMOD modes
        # TODO: set up variables for our CHMOD (WIP)
        self.chmod_modes = ("ISUID", "ISGID", "ENFMT", "ISVTX", "IREAD", "IWRITE", "IEXEC", "IRWXU", "IWUSR", "IXUSR"
                            "IRWXG", "IRGRP", "IWGRP", "IXGRP", "IRWXO", "IROTH", "IWOTH", "IXOTH")
        # chmod description dictionary
        self.chmod_descriptions = {"ISUID": "Set user ID on execution.", "ISGID": "Set group ID on execution.",
                                   "ENFMT": "Record locking enforced.",  "ISVTX": "Save text image after execution.",
                                   "IREAD": "Read by owner.",            "IWRITE": "Write by owner.",
                                   "IEXEC": "Execute by owner.",         "IRWXU": "Read, write, and execute by owner.",
                                   "IRUSR": "Read by owner.",            "IWUSR": "Write by owner.",
                                   "IXUSR": "Execute by owner.",         "IRWXG": "Read, write, and execute by group.",
                                   "IRGRP": "Read by group.",            "IWGRP": "Write by group.",
                                   "IXGRP": "Execute by group.",         "IRWXO": "Read, write, and execute by others.",
                                   "IROTH": "Read by others.",           "IWOTH": "Write by others.",
                                   "IXOTH": "Execute by others."}

        # if silent is false
        if not silent:
            # second stdout.write and flush to clear old message
            # and replace it with a new one
            sys.stdout.write(f"\r{self.class_name} Initialized \n")
            sys.stdout.flush()

    def is_directory(self, directory, silent=False):
        # setup help() return for the method
        """
        method to return True or False as to whether a directory exists

        :param directory: path to the directory the function will check for

        :param silent: a True or False bool to determine whether or not to make print statements

        :return: has 2 possible returns (True, False)
                 True = directory exists
                 False = directory does not exist
        """

        # establish method name for print/debug statements
        method_name = f"{self.class_name}.is_directory()"

        # load the return of os.path.isdir() into a variable
        # to later be returned as True or False
        try:
            exists = os.path.isdir(directory)
        # account for invalid input in directory
        except TypeError:
            if not silent:
                print(f"{method_name}: (({os.getcwd()}\\<{directory}>)) triggered a TypeError in is_directory()")
                print("check your input")
            # and give the exists bool to False
            exists = False

        # if silent is false
        if not silent:
            # and the directory exists
            if exists:
                # say found
                print(f"{method_name}: ({os.getcwd()}\\<{directory}>) found")
            # and if the file does not exist
            else:
                # say 'I'm sorry for your loss' -- I mean -- "not found"
                print(f"{method_name}: ({os.getcwd()}\\<{directory}>) not found")

        # return the True False nature of our exists variable
        return exists

    # WORKING
    # method to return True or False as to whether a file exists
    #
    # has 2 possible returns (True, False)
    # # True = file exists
    # # False = file does not exist
    #
    def is_file(self, filename, silent=False):
        # setup help() return for the method
        """
        method to return True or False as to whether a file exists

        :param filename: path to file the function will check for

        :param silent: a True or False bool to determine whether or not to make print statements

        :return: has 2 possible returns (True, False)
                 True = file exists
                 False = file does not exist
        """

        # establish method name for print/debug statements
        method_name = f"{self.class_name}.is_file()"

        # load the True/False os.path.isfile() return into our exists variable
        # to later be returned as a file exists or file doesnt exist
        exists = os.path.isfile(filename)

        # if we can blab
        if not silent:
            # and the file exists
            if exists:
                # tell our super fans that files are found, and that
                # they are indeed real
                # -- just like ufo's --
                print(f"{method_name}: ({os.getcwd()}\\<{filename}>) found")
            # or if the file doesn't exist
            else:
                # file not found
                print(f"{method_name}: ({os.getcwd()}\\<{filename}>) not found")

        # return the True False nature of our exists variable
        return exists

    # # TODO: MAKE CHANGE DIRECTORY METHOD (ON HOLD)
    # def change_directory(self, destination):
    #     pass
    #
    # # TODO: MAKE CHMOD METHOD (ON HOLD)
    # def chmod(self, filename, mode, silent=False):
    #     """
    #     help test
    #     """
    #
    #     # CHMOD MODES
    #     if mode == "help":
    #         print(f"{self.chmod_descriptions}")
    #
    #     if mode in self.chmod_modes:
    #         _mode = None
    #
    #         # Set user ID on execution
    #         if mode == "ISUID":
    #             _mode = stat.S_ISUID
    #
    #         # Set group ID on execution
    #         elif mode == "ISGID":
    #             _mode = stat.S_ISGID
    #
    #         # Record locking enforced
    #         elif mode == "ENFMT":
    #             _mode = stat.S_ENFMT
    #
    #         # Save text image after execution
    #         elif mode == "ISVTX":
    #             _mode = stat.S_ISVTX
    #
    #         # Read by owner
    #         elif mode == "IREAD":
    #             _mode = stat.S_IREAD
    #
    #         # Write by owner
    #         elif mode == "IWRITE":
    #             _mode = stat.S_IWRITE
    #
    #         # Execute by owner
    #         elif mode == "IEXEC":
    #             _mode = stat.S_IEXEC
    #
    #         # Read, write, and execute by owner
    #         elif mode == "IRWXU":
    #             _mode = stat.S_IRWXU
    #
    #         # Read by owner
    #         elif mode == "IRUSR":
    #             _mode = stat.S_IRUSR
    #
    #         # Write by owner
    #         elif mode == "IWUSR":
    #             _mode = stat.S_IWUSR
    #
    #         # Execute by owner
    #         elif mode == "IXUSR":
    #             _mode = stat.S_IXUSR
    #
    #         # Read, write, and execute by group
    #         elif mode == "IRWXG":
    #             _mode = stat.S_IRWXG
    #
    #         # Read by group
    #         elif mode == "IRGRP":
    #             _mode = stat.S_IRGRP
    #
    #         # Write by group
    #         elif mode == "IWGRP":
    #             _mode = stat.S_IWGRP
    #
    #         # Execute by group
    #         elif mode == "IXGRP":
    #             _mode = stat.S_IXGRP
    #
    #         # Read, write, an execute by others
    #         elif mode == "IRWXO":
    #             _mode = stat.S_IRWXO
    #
    #         # Read by others
    #         elif mode == "IROTH":
    #             _mode = stat.S_IROTH
    #
    #         # Write by others
    #         elif mode == "IWOTH":
    #             _mode = stat.S_IWOTH
    #
    #         # Execute by others
    #         elif mode == "IXOTH":
    #             _mode = stat.S_IXOTH
    #
    #         # call os.chmod() to carry out our action
    #         os.chmod(filename, _mode)
    #
    #     # TODO: FIND A WAY TO CHECK CHMOD AND VERIFY CHANGE
    #
    #     else:
    #         if not silent:
    #             print(f"{method_name}: CHMOD mode out of range")
    #             for key, value in self.chmod_descriptions.items():
    #                 print(f"{key}: {value}")

    def split_filename(self, filename, silent=False):
        # setup help() return for the method
        """
        method to split directories and file names out of file paths and return both directory and file name

        :param filename: path to the file the function will split the filename and directory from
        :param silent: a True or False bool to determine whether or not to make print statements
        :return: returns the post split filename on index[0] and directory on index[1]
        """

        # establish method name for print/debug statements
        method_name = f"{self.class_name}.split_filename()"

        # if the filename contains "/"s which are a good indicator
        # that they hold a directory path
        if "/" in filename:
            # split the directory into a list separated by /'s
            new_directory = filename.split("/")
            # create a [:-value] location for the beginning of our actual filename
            # by subtracting the length of the last split against he original length
            # of the filename
            extract_index = len(filename) - len(new_directory[-1])
            directory = filename[:extract_index]
            # declare our new filename as the file index in new_directory
            filename = new_directory[-1]

        # other wise there is no directory hidden in filename, and we can just leave things as they are
        else:
            # and assign a None to directory
            directory = None

        # if not silent
        if not silent:
            # we'll print out the returns
            print(f"{method_name}:")
            print(f"working directory: {os.getcwd()}")
            print(f"directory: {directory}")
            print(f"filename: {directory}")

        # and/or return the filename, and directory
        return filename, directory

    def touch_file(self, filename, directory=None, mode=None, silent=False):
        # setup help() return for the method
        """
        method to create a file (and associated directories if needed) if it does
        not already exist.

        :param filename:  path to the file (may include directory) the function will create

        :param directory: OPTIONAL path to the directory containing the file

        :param mode: has 2 possible modes "w" and "wb"
                     "w" to touch a plain-text file into existence
                     "wb" to touch a bytes file into existence


        :param silent: a True or False bool to determine whether or not to make print statements

        :return: has 2 possible returns (True, False)
                 True: infers that the file exists and that we were otherwise successful
                 False: infers that the file not only did not exists, but continues not existing
                        and that the method has failed

        """

        # establish method name for print/debug statements
        method_name = f"{self.class_name}.touch_file()"

        # check if there is a value in the directory field
        # if not we want to double check against the filename
        if directory is None:
            # call our split_filename() method and check index 0 for a filename
            # and index 1 for a directory and load them into temporary variables
            # to preserve our original filename
            temp_filename, temp_directory = self.split_filename(filename, silent=True)

            # we make sure that the directory that split_filename() method
            # returned isn't None, meaning it has a value, and therefore the
            # original filename had a directory in it.
            if temp_directory is not None:
                # make sure the filename is updated to reflect the split
                filename = temp_filename
                # make sure the directory is updated to reflect the split
                directory = temp_directory

                # next we want to make sure that the directory does or does not already exist
                # and create it if it doesn't using our make_directory() method
                self.make_directory(directory, silent=True)

            # no hidden directory so pass out to next phase
            else:
                pass

        # if there is a value in the field
        else:
            # if the directory doesn't exist, create it
            # our make_directory() function automatically checks
            # check if the directory exists first
            self.make_directory(directory, silent=True)

        # if there is some form of value in directory at this point
        # we should append the directory and filename together
        # at this point either way they should be split
        if directory is not None:
            filename = f"{directory}{filename}"

        # set the proper modes and content string type based on mode field argument
        # write
        if mode in self.write_modes:
            mode = "w"
            blank_content = ""

        # write bytes
        elif mode in self.write_bytes_modes:
            mode = "wb"
            # use a bytes string for write bytes
            blank_content = b""

        # if mode is None or w/e just default to write mode
        else:
            mode = "w"
            blank_content = ""

        # if the file doesn't exist
        if not self.is_file(filename, silent=True):
            # protect from errors -- shouldn't get them -- but we're protected if we do.
            try:
                # open it with a context manager as f and write an empty blank line based on our selected mode
                # inserting the respective string type as blank_content
                # - the context manager will automatically close the file for us <3
                with open(filename, mode) as f:
                    f.write(blank_content)
            # take care of all those pesky potential errors due to potential user error
            # and return false to declare method failure
            except ValueError:
                if not silent:
                    print(f"{method_name}: ValueError with ('{filename}') as the filename, running in '{mode}' mode")
                return False

            except TypeError:
                if not silent:
                    print(f"{method_name}: TypeError with ('{filename}') as the filename, running in '{mode}' mode")
                return False

            except FileNotFoundError:
                if not silent:
                    print(f"{method_name}: FileNotFoundError with  ('{filename}') "
                          f"as the filename, running in '{mode}' mode")
                return False

            # load the bool of is_file() into a variable to determine whether or not we failed to touch the file
            exists = self.is_file(filename, silent=True)
            # if the file exists, we succeeded
            if exists:
                # if we're chatty cathy
                if not silent:
                    # say "we rule, our genes are on top of shit, we won, everything is good and nothing is bad"
                    print(f"{method_name}: successfully touched ({os.getcwd()}\\<{filename}>)")

            # if the file doesn't exist
            else:
                # and we're talking
                if not silent:
                    # say "we suck, and failed a simple task... that we are failures -- just like our ancestors"
                    # let them know that this failure is not only a programmatic failure but also a genetic one
                    print(f"{method_name}: failed to touch ({os.getcwd()}\\<{filename}>)")

            # and/or return the exists bool to represent success or failure
            return exists

        # if the file already exists, skip the whole song and dance and write home about it
        else:
            # if we can talk
            if not silent:
                # we will talk, we'll tell them how their precious file already exists
                print(f"{method_name}: ({os.getcwd()}\\<{filename}>) already exists")

                # we will return a True value because the file existing, is functionally
                # the same as successfully touching the file
                return True

    def make_directory(self, directory, silent=False):
        # setup help() return for the method
        """
        method to CREATE NEW DIRECTORIES if they don't already exists

        :param directory: path to the directory the function is creating

        :param silent: a True or False bool to determine whether or not to make print statements

        :return: has 2 possible return values of True and False (bool)
                 True: The directory was successfully created, or otherwise already exists
                       both are functionally the same
                 False: The directory does not exist, or the method was otherwise unable to
                        create the directory
        """

        # establish method name for print/debug statements
        method_name = f"{self.class_name}.make_directory()"

        # if the file exists pass
        if self.is_directory(directory, silent=True):
            # if silent == false
            if not silent:
                # let them know that the directory exists and that there is no need
                # to create it
                print(f"{method_name}: could not create the directory ({os.getcwd()}\\<{directory}>)")
                print("because it already exists")

            # we'll return with a True, to say that YES the directory is there
            # even though *this method* didn't create it.
            # the goal of having the directory exist is still met.
            return True

        # otherwise create the directory
        else:
            # attempt to create the directory using os.makedirs()
            try:
                os.makedirs(directory)
            # if TypeError
            except TypeError:
                if not silent:
                    print(f"{method_name}: ({os.getcwd()}\\<{directory}>) triggered a TypeError in "
                          f"make_directory({directory})")
                    print("check your input")
                # return False to state that the method failed
                return False

            # if ValueError
            except ValueError:
                if not silent:
                    print(f"{method_name}: ({os.getcwd()}\\<{directory}>) triggered a ValueError in "
                          f"make_directory({directory})")
                    print("check your input")
                # return False to state that the method failed
                return False
            # check if the directories were successfully created and return the bool into a variable
            exists = self.is_directory(directory, silent)
            # use the exists bool to determine how we continue
            # if directories were successfully created
            if exists:
                # if silent == false
                if not silent:
                    # let them know that it was a success
                    print(f"{method_name}: ({os.getcwd()}\\<{directory}>) was successfully created!")
            # if directories were unsuccessfully created
            else:
                # if silent == false
                if not silent:
                    # let them know that it was a failure
                    print(f"{method_name}: failed to create ({os.getcwd()}\\<{directory}>) as a directory")

            # return our bool to determine success or failure
            return exists

    def read_file(self, filename, mode=None, silent=False):
        # setup help() return for the method
        """
        method to READ return the contents of a file

        :param filename: path to the file the function is reading

        :param mode: has 3 possible modes ('r', 'rb', 'None')
                     'None' - Attempts to decide the read mode automatically (Experimental)
                     'r'    - reads the file in 'r' (plain text) mode
                     'rb'   - reads the file in 'rb' (bytes) mode

        :param silent: a True or False bool to determine whether or not to make print statements

        :return: has 2 possible returns (Content, None)
                 the method should return the content of the file
                 or if an error is encountered along the way
                 it will return None
        """

        # establish method name for print/debug statements
        method_name = f"{self.class_name}.read_file()"

        if self.is_file(filename, silent=True):
            # set mode accordingly
            # TODO: Investigate additional read modes
            # read plain
            if mode in self.read_modes:
                mode = "r"
            # read bytes
            elif mode in self.read_bytes_modes:
                mode = "rb"

            # if the mode is None / unfilled
            elif mode is None:
                #       INCLUDE AUTOMATIC PROCESSING DURING "mode=None" CALLS
                #       -- ENDED UP TAKING A DIFFERENT APPROACH, WE USE TRY/EXCEPT TYPEERROR
                #          TO CATCH IF WE HAVE THE WRONG MODE, IF WE DO WE TRY THE OTHER READ
                #          MODE.
                #          HOPING THIS WORKS!
                #       -- (SEEMS TO BE WORKING)

                try:
                    with open(filename, "r") as f:
                        return f.read()
                except TypeError:
                    with open(filename, "rb") as f:
                        return f.read()

            # use a context manager to open our file and return its contents
            # the context manager will automatically close the file once
            try:
                with open(filename, mode) as f:
                    return f.read()
            # handle pesky errors that may arise for whatever reason and inform the user
            # and then pass out of the method, returning None
            except ValueError:
                if not silent:
                    print(f"{method_name}: ValueError with ('{filename}') as the filename, running in '{mode}' mode")
                return None
            except TypeError:
                if not silent:
                    print(f"{method_name}: TypeError with ('{filename}') as the filename, running in '{mode}' mode")
                return None

        # if the file doesn't
        else:
            # if we can tell secrets
            if not silent:
                # announce it to the world
                print(f"file not found.")
            # and return None to represent the file doesn't exist
            return None

    # TODO: DONE?
    def write_file(self, content, filename, mode=None, encoding='utf-8', silent=False):
        # setup help() return for the method
        """
        method to WRITE FILES

        :param content: content you want to write to the file

        :param filename: path to the file the function is writing

        :param mode: write mode, comes in two 5 flavors, write, writebytes, append, appendbytes, and None
                     None(Default) mode - will determine if the file exists and make it if it doesn't
                                          and determine from the content type whether or not to write
                                          in bytes or plain modes.  --  this is essentially an automatic
                                          mode.

                     "w" (write) mode - will write the file as a plain text with a new line at the end
                                        !CAUTION! - THIS WILL WIPE THE FILE'S PREVIOUS CONTENTS, IF ANY
                                                    AND REPLACE IT WITH THE CONTENTS OF THE CONTENT FIELD

                     "wb" (write bytes) mode - this mode is a lot like "w" mode except it writes in bytes
                                               without a new line, but carries the same content wiping
                                               warning as "w" mode.

                     "a" (append) mode - this mode should be used on existing files, and will add the "content" field
                                         to the last line of a plain text type file (such as .txt)

                     "ab" (append bytes) mode - this mode is meant to add bytes type content directly to the end
                                                of a bytes file (such as .jpg, or other data type) :shrug:

        :param encoding: encoding used for bytes mode -- example usage: 'utf-8'

        :param silent: a True or False bool to determine whether or not to make print statements

        :return: (True, False) bool return to confirm that the file exists at end of function
                 if it returns false it means that the write has somehow failed.
        """

        # establish method name for print/debug statements
        method_name = f"{self.class_name}.write_file()"

        # we'll set our modes to their proper values based on the user's call
        # and assign the proper content type
        # set mode to standard write
        if mode in self.write_modes:
            mode = "w"
            # create a new line each entry
            content = f'{content}\n'

        # set mode to write bytes
        elif mode in self.write_bytes_modes:
            mode = "wb"
            # make sure our data is bytes type and give it an encoding
            content = bytes(content, encoding)

        # set mode to append standard
        elif mode in self.write_append_modes:
            mode = "a"
            # create a new line each entry
            content = f'{content}\n'

        # set mode to append bytes
        elif mode in self.write_append_bytes_modes:
            mode = "ab"
            # make sure our data is bytes type and give it an encoding
            content = bytes(content, encoding)

        # if the mode field argument is empty
        if mode is None:
            # if the file exists
            if self.is_file(filename, silent=True):
                # and is_bytes is set to zero
                if isinstance(content, bytes):
                    # set to append bytes
                    mode = "ab"
                # if the content isn't bytes
                else:
                    # set to append (plain txt)
                    mode = "a"
                    # set content to string() type
                    content = f'{content}\n'
            # if the file doesnt exist
            else:
                # and is_bytes is set to zero
                if isinstance(content, bytes):
                    # set to write bytes
                    mode = "wb"
                # if is_bytes is anything but 0
                else:
                    # set to write (plain txt)
                    mode = "w"
                    # set content to string() type
                    content = f'{content}\n'

        # let's make sure that we don't get any errors
        try:
            # with the context manager, open the file and write our content
            # the context manager will automatically close the file once
            # it is finished
            # pass in our filename and the filtered out mode
            with open(filename, mode) as f:
                # write the content to the file
                f.write(content)
        # such as value errors or...
        except ValueError:
            if not silent:
                print(f"{method_name}: ValueError with ('{filename}') as the filename, running in '{mode}' mode")
            return False
        # type errors
        except TypeError:
            if not silent:
                print(f"{method_name}: TypeError with ('{filename}') as the filename, running in '{mode}' mode")
            return False

        # after we finish our write we can return whether or not the file
        # was successfully created using our is_file() method
        exists = self.is_file(filename, silent=True)

        # if our file is being written in w or wb modes
        # test to make sure the file was created
        if mode == "w" or mode == "wb":
            # if the file exists
            if exists:
                if not silent:
                    print(f"{method_name}: successfully wrote {filename} in '{mode}' mode")
            else:
                if not silent:
                    print(f"{method_name}: failed to write {filename} in '{mode}' mode")
        return exists

    def remove_directory(self, directory, mode="single", silent=False):
        # setup help() return for the method
        """
        method to REMOVE DIRECTORIES

        :param directory: path to the directory the function is removing

        :param mode: method has 3 modes ("single", "all", and "clear")
                    "single" mode - removes the end point target directory, ( will not remove files )

                    "all" mode - removes all directories between the working directory
                                 and the target directory. ( will not remove files)

                    "clear" mode - removes a directory and all of its children
                                   !! CAUTION YOU WILL LOSE ALL CONTENT UNDER THE TARGET DIRECTORY!!


        :param silent: a True or False bool to determine whether or not to make print statements

        :return: the method has 3 possible returns
                 (True, False, None)
                 True = Directory Removed
                 False = Directory Removal Failed
                 None = Directory Does Not Exist

        """

        # establish method name for print/debug statements
        method_name = f"{self.class_name}.remove_directory()"

        # we'll just take a bit of code from our remove_file() method and turn it into
        # a reusable sub method to confirm the success of the main method, across modes
        def confirm_remove(sf_directory, sf_silent):
            # use our is_directory to determine whether or not we successfully deleted the directories
            # and return a bool for success/failure
            _exists = self.is_directory(sf_directory, silent=True)
            if _exists:
                # if silent is false make announcements
                if not sf_silent:
                    print(f"{method_name}: ({os.getcwd()}\\<{sf_directory}>) unsuccessfully removed")

            # if the file still exists the function has failed remove it
            else:
                # if silent is false
                if not sf_silent:
                    # make announcements
                    print(f"{method_name}: ({os.getcwd()}\\<{sf_directory}>) successfully removed")
            # return the bool exists to determine success or failure
            return _exists

        # begin method logic
        # check if the directory exists
        if self.is_directory(directory, silent=True):
            # and filter modes between single and all

            # "single" mode deletes the end directory only.
            if mode == "single":
                # make sure that the file directory is empty
                try:
                    # use the os.rmdir() function to remove our directory
                    os.rmdir(directory)

                    # return whether or not the directory was deleted
                    # by using our sub method
                    return confirm_remove(directory, silent)

                # if we except with an OSError, it means that the directory has contents
                except OSError:
                    # if silent set to !FALSE! ( ͡ʘ ͜ʖ ͡ʘ) *slowly unwinds*
                    if not silent:
                        # blah blah lah
                        print(f"{method_name}: ({os.getcwd()}\\<{directory}>) is not empty")
                        print(f"Consider using 'clear' mode to remove the directory, all sub-directories, and files")

                    # and/or return as False, directory could not be removed.
                    return False

            # "all" mode removes all directories between the working directory and the target directory
            elif mode == "all":
                # make sure that the file directory is empty with try/except statement
                try:
                    # call os.removedirs() with our directory passed in to remove all
                    # directories between working and target
                    os.removedirs(directory)

                    # return whether or not the directory was deleted
                    # by using our sub method
                    return confirm_remove(directory, silent)

                # if we except with an OSError, it means that the directory has contents
                except OSError:
                    # if silent is false
                    if not silent:
                        # squaaawwwk!!
                        print(f"{method_name}: ({os.getcwd()}\\<{directory}>) is not empty")
                        print(f"Consider using 'clear' mode to remove the directory, all sub-directories, and files")

                    # and/or return as False, directory could not be removed.
                    return False

            # "clear" mode removes everything under a target directory
            elif mode == "clear":
                # using the shutil.rmtree(directory) method.
                shutil.rmtree(directory)

                # let's check that our directory has been deleted
                exists = self.is_directory(directory, silent=True)
                # FAILURE
                if exists:
                    if not silent:
                        print(f"{method_name}: could not clear ({os.getcwd()}\\<{directory}>)")

                # SUCCESS
                else:
                    if not silent:
                        print(f"{method_name}: successfully cleared ({os.getcwd()}\\<{directory}>) and "
                              f"all sub-directories, and files")

                # we return the True False bool of exists to indicate success or failure.
                return exists

        # if the directory doesn't exist
        else:
            # if not silent
            if not silent:
                # let the user know that the directory doesn't exist.
                print(f"{method_name}: could not remove ({os.getcwd()}\\<{directory}>)")
                print(f"directory does not exist!")

            # return that the directory is None (Does Not Exist)
            return None

    # TODO: add secure delete with some sort of looped overwrite function?  Need to research before diving
    #       into this one.
    def remove_file(self, filename, silent=False):
        # setup help() return for the method
        """
        method to REMOVE FILES

        :param filename: path to the file the function is removing

        :param silent: a True or False bool to determine whether or not to make print statements

        :return: has 3 possible returns (True, False, None)
                 (True, False, None)
                 True = File Removed
                 False = File Removal Failed
                 File Does Not Exist
        """

        # establish method name for print/debug statements
        method_name = f"{self.class_name}.remove_file()"

        # silently make sure the file exists
        if self.is_file(filename, silent=True):
            # if the file exists use the imported os module to remove it
            os.remove(filename)

            # use our is_file to determine whether or not we successfully deleted the file
            # and return a bool for success/failure
            if not self.is_file(filename, silent=True):
                # if silent is false
                if not silent:
                    # make announcements
                    print(f"{method_name}: ({os.getcwd()}\\<{filename}>) successfully removed")

                # and/or return that we have removed the file
                return True

            # if the file still exists the function has failed remove it
            else:
                # if silent is false make announcements
                if not silent:
                    print(f"{method_name}: ({os.getcwd()}\\<{filename}>) unsuccessfully removed")

                # and/or return that we have failed to remove the file
                return False

        # if the file doesn't exist...
        else:
            # if not silent, let the user know that the file doesn't exist.
            if not silent:
                # if the file doesnt exist, then run away screaming and crying
                print(f"{method_name}: could not remove ({os.getcwd()}\\<{filename}>)")
                print(f"file does not exist.")

            # and/or return that the file is None (Does Not Exist)
            return None


class PrintHandler:

    def __init__(self, verbosity=0):
        # setup help() return for the class
        """
        WORK IN PROGRESS

        PrintHandler Class Initializer method.
        establishes various class wide variables

        :param verbosity: control the print level when log_mode is set
                          0 - Print all debug messages - *Default
                          1 - Print on Info -> Error only
                          2 - Print on Debug -> Error only
                          3 - Print on Warning -> Error only
                          4 - Print on Error only
        """

        # establish class name for print/debug statements
        self.class_name = f"PrintHandler()"
        self.verbosity = verbosity

        try:
            import colorama
            # initialize colorama
            self.colorama = colorama
            self.colorama.init()
            print(f"{self.class_name} Initialized")
        except ImportError:
            print(f"{self.class_name} Initialization Failed!")
            print(f"Please install colorama with 'pip install colorama' in your terminal")

    def write_log(self, message=None, logfile=None):
        # setup help() return for the method
        """
        method to write messages to a logfile

        :param message: message you want to log
        :param logfile: path to the file you want to act as your logfile
        :return: has no return
        """

        # establish method name for print/debug statements
        method_name = f"{self.class_name}.write_log()"

        # get the current time
        now = str(datetime.datetime.now())

        # if there is no message skip the nonsense
        if message is None:
            pass
        # if there is a message to write, continue
        else:
            # if the log file is empty, default to "logfile.txt"
            if logfile is None:
                logfile = "logfile.txt"

            # otherwise let us check if the file exists
            if os.path.isfile(logfile):
                # if it does we'll open the file and make an append entry
                with open(logfile, "a") as f:
                    f.write(f"({now[:-7]})  - {message}\n")

            # if the file doesn't exist we'll make it with our first write entry
            else:
                with open(logfile, "w") as f:
                    f.write(f"({now[:-7]})  - {message}\n")

    # method to act as a custom print feature
    def c_print(self, message=None, sep=None, end=None, file=None, flush=False,
                color=None, timestamp=False, log_mode=None):
        # setup help() return for the method
        """
        method to act as a custom print feature

        :param message: the string that you want to push to console

        :param sep: separator in print()
        :param end: end of string in print()
        :param file: file in print()
        :param flush: flush in print()

        :param color: color setting of print
                      BLACK, RED, L_RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE

        :param timestamp: whether or not to print with a timestamp

        :param log_mode: what warning level to print in if any
                         0 - None
                         1 - INFO - CYAN
                         2 - DEBUG - YELLOW
                         3 - WARNING - L_RED
                         4 - ERROR - RED

        :return: has no return
        """

        # establish method name for print/debug statements
        method_name = f"{self.class_name}.c_print()"

        # verbosity level of print (log_mode)
        log_modes = (0, 1, 2, 3, 4)
        if log_mode in log_modes:
            if log_mode == 0:
                log_mode = None

            # INFO
            if log_mode == 1:
                color = "CYAN"
                message = f"INFO: {message}"

            # DEBUG
            if log_mode == 2:
                color = "YELLOW"
                message = f"DEBUG: {message}"

            # WARNING
            if log_mode == 3:
                color = "L_RED"
                message = f"WARNING: {message}"

            # ERROR
            if log_mode == 4:
                color = "RED"
                message = f"ERROR: {message}"

        # if our log_mode is out of range, just set it to None
        else:
            log_mode = None

        # if our timestamp is set to True add a datetime.datetime.now stamp to our message
        if timestamp:
            # datetime.datetime.now() is not subscriptable, so we'll convert it into a string
            # to where we can slice it with indexing to erase the unwanted tail at the end
            now = str(datetime.datetime.now())
            message = f"({now[:-7]}) - {message}"

        # do not print log statements lower than the settings.py allows and the log_mode is acceptable
        do_not_print = False
        if log_mode in log_modes and self.verbosity > log_mode:
            do_not_print = True

        # check if we need to skip the print and pass out of the function
        if do_not_print:
            pass

        # otherwise we continue
        else:
            # list of acceptable colors
            colors = ("BLACK", "RED", "L_RED", "GREEN", "YELLOW", "BLUE", "MAGENTA", "CYAN", "WHITE")
            # if the color isn't in the list of colors or is otherwise empty
            if color is None or color not in colors:
                # make sure we print a freshly reset message as to remove any previous formatting from
                # previous calls
                print(f"{self.colorama.Fore.RESET}{message}",
                      sep=sep, end=end, file=file, flush=flush)

            # if our log_mode is None or it is acceptable, we assign colors to the message
            if log_mode is None or log_mode in log_modes:
                # write log_mode messages to a log file
                if log_mode in log_modes and log_mode is not None:
                    if log_mode >= self.verbosity:
                        self.write_log(message)

                # add color to and print the message, and close
                # with a Fore.RESET to prevent color settings from bleeding
                # into subsequently printed statements
                if color == "BLACK":
                    print(f"{self.colorama.Fore.WHITE}{message}{self.colorama.Fore.RESET}",
                          sep=sep, end=end, file=file, flush=flush)
                if color == "L_RED":
                    print(f"{self.colorama.Fore.LIGHTRED_EX}{message}{self.colorama.Fore.RESET}",
                          sep=sep, end=end, file=file, flush=flush)
                if color == "RED":
                    print(f"{self.colorama.Fore.RED}{message}{self.colorama.Fore.RESET}",
                          sep=sep, end=end, file=file, flush=flush)
                if color == "GREEN":
                    print(f"{self.colorama.Fore.GREEN}{message}{self.colorama.Fore.RESET}",
                          sep=sep, end=end, file=file, flush=flush)
                if color == "YELLOW":
                    print(f"{self.colorama.Fore.YELLOW}{message}{self.colorama.Fore.RESET}",
                          sep=sep, end=end, file=file, flush=flush)
                if color == "BLUE":
                    print(f"{self.colorama.Fore.BLUE}{message}{self.colorama.Fore.RESET}",
                          sep=sep, end=end, file=file, flush=flush)
                if color == "MAGENTA":
                    print(f"{self.colorama.Fore.MAGENTA}{message}{self.colorama.Fore.RESET}",
                          sep=sep, end=end, file=file, flush=flush)
                if color == "CYAN":
                    print(f"{self.colorama.Fore.CYAN}{message}{self.colorama.Fore.RESET}",
                          sep=sep, end=end, file=file, flush=flush)
                if color == "WHITE":
                    print(f"{self.colorama.Fore.BLACK}{message}{self.colorama.Fore.RESET}",
                          sep=sep, end=end, file=file, flush=flush)


#
# If you're reading this...
# Hi!  I'm Billy B. Bobby Bobertson Jr. III. -- my friend's call me Bob...
# I hope you are well, and are enjoying yourself wherever you may be. <3
#
# And If you are not reading this... ^^ this ^^ to you ALL the same (@everyone).
#


if __name__ == '__main__':
    print("This file is part of the junkyard pip module and isn't intended to be run from source")
    # TODO: GET FILES READY FOR SECOND PIP COMMIT! (0.0.03a)

    # TODO: FOR SECOND PIP COMMIT # (0.0.03a)
    #   - GENERAL CHANGES (WIP)
    #       - REWORK ALL HELP TAGS (DONE!)
    #       - REWORK self.class_name USAGE and ADD method_name's to ALL METHODS (DONE!)
    #       - EXPAND USAGE OF DEBUG PRINT STATEMENTS TO GIVE ALL OF OUR method_name VARIABLES USE? (ON HOLD)
    #   - FILEHANDLER()
    #       - ADD ABILITY TO GET AND CHANGE WORKING DIRECTORY!
    #       - ADD ABILITY TO COPY FILES / DIRECTORIES?
    #       - ADD ABILITY TO MOVE FILES / DIRECTORIES?
    #       - ADD ABILITY TO CHMOD?
    #       - REWORK ALL WRITE CALLS TO INCLUDE TEMP FILE
    #         TO PREVENT ANY POTENTIAL CORRUPTION!
    #   - ADD A DIAGNOSTICSHANDLER() CLASS (DONE!)
    #       - INCORPORATE SPEED TESTERS INTO THE NEW CLASS (DONE!)
    #   - ADD A PRINTHANDLER() CLASS (DONE!)
    #       - INCORPORATE C_PRINT AND LOGGER INTO NEW CLASS (DONE!)
    #   - ADD AN ENCRYPTION HANDLER?
    #       - GIVE CRYPTOHANDLER() LIFE / FUNCTIONALITY
    #       - INCORPORATE CRYPTOHANDLER() into FILEHANDLER() ???



# let's do a splash screen when our module is imported and initialized
# feel free to change this
splash = True

module_name = "Junkyard"
__version__ = "0.0.3a5"

# if not main file
if __name__ != "__main__":
    # print a version of the
    if splash:
        print(f"Powered by {module_name} ({__version__})")



import multiprocessing
"""
THIS LIBRARY IS CURRENTLY EMPTY AND ONLY CONTAINS PLACEHOLDER CLASSES AND METHODS

IT IS INTENDED TO BE THE ROBOTICS AND ARTIFICIAL INTELLIGENCE SECTION OF THE JUNKYARD
IN THE NEAR TO NEVER-REALIZED FUTURE.
"""

# inform the user that this file is useless for the time being
if __name__ != "__main__":
    print("The Cyberdyne module has yet to be fleshed out and currently contains zero functionality.")
    print("It is our hope that this will change in the near future -- Until then this file is useless")


# a class to act as the cyberdyne library's core class
class SkyNetCore:
    # SkyNetCore() initialization method
    def __init__(self, name, designation, model=None):
        # give the core a personalized name
        self.name = name
        # give the core a designation
        self.designation = designation
        # establish our ai model
        self.model = model
        # get threads in a cpu
        self.cpu_thread_count = multiprocessing.cpu_count()


if __name__ == '__main__':
    # test SkyNetCore
    snc = SkyNetCore("Ro-Bob", "B-DROID-01A")
